@@ -107,14 +107,10 @@ calc_fractal = calc_fractal_opencl
 # calc_fractal = calc_fractal_numpy
 
 if __name__ == '__main__':
-    try:
-        import Tkinter as tk
-    except ImportError:
-        # Python 3
-        import tkinter as tk
+    import tkinter as tk
     from PIL import Image, ImageTk
 
-    class Mandelbrot(object):
+    class Mandelbrot:
         def __init__(self):
             # create window
             self.root = tk.Tk()
