@@ -1,8 +1,10 @@
-"""Scan primitive."""
+"""Algorithms built on scans."""
 
-from __future__ import division
 
-__copyright__ = """Copyright 2011-2012 Andreas Kloeckner"""
+__copyright__ = """
+Copyright 2011-2012 Andreas Kloeckner
+Copyright 2017 Hao Gao
+"""
 
 __license__ = """
 Permission is hereby granted, free of charge, to any person
@@ -27,15 +29,40 @@ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 """
 
+
 import numpy as np
 import pyopencl as cl
 import pyopencl.array  # noqa
 from pyopencl.scan import ScanTemplate
-from pyopencl.tools import dtype_to_ctype
+from pyopencl.tools import dtype_to_ctype, get_arg_offset_adjuster_code
 from pytools import memoize, memoize_method, Record
 from mako.template import Template
 
 
+# {{{ "extra args" handling utility
+
+def _extract_extra_args_types_values(extra_args):
+    from pyopencl.tools import VectorArg, ScalarArg
+
+    extra_args_types = []
+    extra_args_values = []
+    extra_wait_for = []
+    for name, val in extra_args:
+        if isinstance(val, cl.array.Array):
+            extra_args_types.append(VectorArg(val.dtype, name, with_offset=False))
+            extra_args_values.append(val)
+            extra_wait_for.extend(val.events)
+        elif isinstance(val, np.generic):
+            extra_args_types.append(ScalarArg(val.dtype, name))
+            extra_args_values.append(val)
+        else:
+            raise RuntimeError("argument '%d' not understood" % name)
+
+    return tuple(extra_args_types), extra_args_values, extra_wait_for
+
+# }}}
+
+
 # {{{ copy_if
 
 _copy_if_template = ScanTemplate(
@@ -61,7 +88,9 @@ def copy_if(ary, predicate, extra_args=[], preamble="", queue=None, wait_for=Non
     :returns: a tuple *(out, count, event)* where *out* is the output array, *count*
         is an on-device scalar (fetch to host with `count.get()`) indicating
         how many elements satisfied *predicate*, and *event* is a
-        :class:`pyopencl.Event` for dependency management.
+        :class:`pyopencl.Event` for dependency management. *out* is allocated
+        to the same length as *ary*, but only the first *count* entries carry
+        meaning.
 
     .. versionadded:: 2013.1
     """
@@ -70,8 +99,12 @@ def copy_if(ary, predicate, extra_args=[], preamble="", queue=None, wait_for=Non
     else:
         scan_dtype = np.int32
 
-    extra_args_types = tuple((val.dtype, name) for name, val in extra_args)
-    extra_args_values = tuple(val for name, val in extra_args)
+    if wait_for is None:
+        wait_for = []
+
+    extra_args_types, extra_args_values, extra_wait_for = \
+        _extract_extra_args_types_values(extra_args)
+    wait_for = wait_for + extra_wait_for
 
     knl = _copy_if_template.build(ary.context,
             type_aliases=(("scan_t", scan_dtype), ("item_t", ary.dtype)),
@@ -81,9 +114,8 @@ def copy_if(ary, predicate, extra_args=[], preamble="", queue=None, wait_for=Non
     count = ary._new_with_changes(data=None, offset=0,
             shape=(), strides=(), dtype=scan_dtype)
 
-    # **dict is a Py2.5 workaround
     evt = knl(ary, out, count, *extra_args_values,
-            **dict(queue=queue, wait_for=wait_for))
+            queue=queue, wait_for=wait_for)
 
     return out, count, evt
 
@@ -153,8 +185,12 @@ def partition(ary, predicate, extra_args=[], preamble="", queue=None, wait_for=N
     else:
         scan_dtype = np.uint32
 
-    extra_args_types = tuple((val.dtype, name) for name, val in extra_args)
-    extra_args_values = tuple(val for name, val in extra_args)
+    if wait_for is None:
+        wait_for = []
+
+    extra_args_types, extra_args_values, extra_wait_for = \
+            _extract_extra_args_types_values(extra_args)
+    wait_for = wait_for + extra_wait_for
 
     knl = _partition_template.build(
             ary.context,
@@ -167,9 +203,8 @@ def partition(ary, predicate, extra_args=[], preamble="", queue=None, wait_for=N
     count = ary._new_with_changes(data=None, offset=0,
             shape=(), strides=(), dtype=scan_dtype)
 
-    # **dict is a Py2.5 workaround
     evt = knl(ary, out_true, out_false, count, *extra_args_values,
-            **dict(queue=queue, wait_for=wait_for))
+            queue=queue, wait_for=wait_for)
 
     return out_true, out_false, count, evt
 
@@ -222,8 +257,12 @@ def unique(ary, is_equal_expr="a == b", extra_args=[], preamble="",
     else:
         scan_dtype = np.uint32
 
-    extra_args_types = tuple((val.dtype, name) for name, val in extra_args)
-    extra_args_values = tuple(val for name, val in extra_args)
+    if wait_for is None:
+        wait_for = []
+
+    extra_args_types, extra_args_values, extra_wait_for = \
+            _extract_extra_args_types_values(extra_args)
+    wait_for = wait_for + extra_wait_for
 
     knl = _unique_template.build(
             ary.context,
@@ -235,9 +274,8 @@ def unique(ary, is_equal_expr="a == b", extra_args=[], preamble="",
     count = ary._new_with_changes(data=None, offset=0,
             shape=(), strides=(), dtype=scan_dtype)
 
-    # **dict is a Py2.5 workaround
     evt = knl(ary, out, count, *extra_args_values,
-            **dict(queue=queue, wait_for=wait_for))
+            queue=queue, wait_for=wait_for)
 
     return out, count, evt
 
@@ -253,13 +291,13 @@ def to_bin(n):
         digs.append(str(n % 2))
         n >>= 1
 
-    return ''.join(digs[::-1])
+    return "".join(digs[::-1])
 
 
-def _padded_bin(i, l):
+def _padded_bin(i, nbits):
     s = to_bin(i)
-    while len(s) < l:
-        s = '0' + s
+    while len(s) < nbits:
+        s = "0" + s
     return s
 
 
@@ -270,7 +308,7 @@ def _make_sort_scan_type(device, bits, index_dtype):
 
     fields = []
     for mnr in range(2**bits):
-        fields.append(('c%s' % _padded_bin(mnr, bits), index_dtype))
+        fields.append(("c%s" % _padded_bin(mnr, bits), index_dtype))
 
     dtype = np.dtype(fields)
 
@@ -280,6 +318,7 @@ def _make_sort_scan_type(device, bits, index_dtype):
     dtype = get_or_register_dtype(name, dtype)
     return name, dtype, c_decl
 
+
 # {{{ types, helpers preamble
 
 RADIX_SORT_PREAMBLE_TPL = Template(r"""//CL//
@@ -376,15 +415,21 @@ RADIX_SORT_OUTPUT_STMT_TPL = Template(r"""//CL//
 
 # {{{ driver
 
-class RadixSort(object):
+# import hoisted here to be used as a default argument in the constructor
+from pyopencl.scan import GenericScanKernel
+
+
+class RadixSort:
     """Provides a general `radix sort <https://en.wikipedia.org/wiki/Radix_sort>`_
     on the compute device.
 
+    .. seealso:: :class:`pyopencl.bitonic_sort.BitonicSort`
+
     .. versionadded:: 2013.1
     """
     def __init__(self, context, arguments, key_expr, sort_arg_names,
             bits_at_a_time=2, index_dtype=np.int32, key_dtype=np.uint32,
-            options=[]):
+            scan_kernel=GenericScanKernel, options=[]):
         """
         :arg arguments: A string of comma-separated C argument declarations.
             If *arguments* is specified, then *input_expr* must also be
@@ -431,7 +476,7 @@ class RadixSort(object):
 
             boundary_mnr = known_bits + "1" + (self.bits-len(known_bits)-1)*"0"
 
-            return ("((mnr < %s) ? %s : %s)" % (
+            return ("((mnr < {}) ? {} : {})".format(
                 int(boundary_mnr, 2),
                 get_count_branch(known_bits+"0"),
                 get_count_branch(known_bits+"1")))
@@ -452,8 +497,7 @@ class RadixSort(object):
         scan_preamble = preamble \
                 + RADIX_SORT_SCAN_PREAMBLE_TPL.render(**codegen_args)
 
-        from pyopencl.scan import GenericScanKernel
-        self.scan_kernel = GenericScanKernel(
+        self.scan_kernel = scan_kernel(
                 context, scan_dtype,
                 arguments=scan_arguments,
                 input_expr="scan_t_from_value(%s, base_bit, i)" % key_expr,
@@ -498,7 +542,7 @@ class RadixSort(object):
         if allocator is None:
             allocator = args[self.first_array_arg_idx].allocator
 
-        queue = kwargs.pop("allocator", None)
+        queue = kwargs.pop("queue", None)
         if queue is None:
             queue = args[self.first_array_arg_idx].queue
 
@@ -514,7 +558,7 @@ class RadixSort(object):
             scan_args = args + sorted_args + [base_bit]
 
             last_evt = self.scan_kernel(*scan_args,
-                    **dict(queue=queue, wait_for=wait_for))
+                    queue=queue, wait_for=wait_for)
             wait_for = [last_evt]
 
             # substitute sorted
@@ -541,7 +585,9 @@ class RadixSort(object):
 
 _LIST_BUILDER_TEMPLATE = Template("""//CL//
 % if double_support:
+    #if __OPENCL_C_VERSION__ < 120
     #pragma OPENCL EXTENSION cl_khr_fp64: enable
+    #endif
     #define PYOPENCL_DEFINE_CDOUBLE
 % endif
 
@@ -580,8 +626,8 @@ typedef ${index_type} index_type;
 
 #define LIST_ARG_DECL ${user_list_arg_decl}
 #define LIST_ARGS ${user_list_args}
-#define USER_ARG_DECL ${user_arg_decl}
-#define USER_ARGS ${user_args}
+#define USER_ARG_DECL ${user_arg_decl_no_offset}
+#define USER_ARGS ${user_args_no_offset}
 
 // }}}
 
@@ -593,7 +639,8 @@ __kernel
 %if do_not_vectorize:
 __attribute__((reqd_work_group_size(1, 1, 1)))
 %endif
-void ${kernel_name}(${kernel_list_arg_decl} USER_ARG_DECL index_type n)
+void ${kernel_name}(
+    ${kernel_list_arg_decl} ${user_arg_decl_with_offset} index_type n)
 
 {
     %if not do_not_vectorize:
@@ -618,18 +665,30 @@ void ${kernel_name}(${kernel_list_arg_decl} USER_ARG_DECL index_type n)
         %else:
             %for name, dtype in list_names_and_dtypes:
                 %if name not in count_sharing:
-                    index_type plb_${name}_index =
-                        plb_${name}_start_index[i];
+                    index_type plb_${name}_index;
+                    if (plb_${name}_start_index)
+                        %if name in eliminate_empty_output_lists:
+                            plb_${name}_index =
+                                plb_${name}_start_index[
+                                    ${name}_compressed_indices[i]
+                                ];
+                        %else:
+                            plb_${name}_index = plb_${name}_start_index[i];
+                        %endif
+                    else
+                        plb_${name}_index = 0;
                 %endif
             %endfor
         %endif
 
+        ${arg_offset_adjustment}
         generate(${kernel_list_arg_values} USER_ARGS i);
 
         %if is_count_stage:
             %for name, dtype in list_names_and_dtypes:
                 %if name not in count_sharing:
-                    plb_${name}_count[i] = plb_loc_${name}_count;
+                    if (plb_${name}_count)
+                        plb_${name}_count[i] = plb_loc_${name}_count;
                 %endif
             %endfor
         %endif
@@ -698,11 +757,14 @@ class ListOfListsBuilder:
     List entries are generated by calls to `APPEND_<list name>(value)`.
     Multiple lists may be generated at once.
 
+    .. automethod:: __init__
+    .. automethod:: __call__
     """
     def __init__(self, context, list_names_and_dtypes, generate_template,
             arg_decls, count_sharing=None, devices=None,
             name_prefix="plb_build_list", options=[], preamble="",
-            debug=False, complex_kernel=False):
+            debug=False, complex_kernel=False,
+            eliminate_empty_output_lists=[]):
         """
         :arg context: A :class:`pyopencl.Context`.
         :arg list_names_and_dtypes: a list of `(name, dtype)` tuples
@@ -717,6 +779,8 @@ class ListOfListsBuilder:
         :arg options: OpenCL compilation options for kernels using
             *generate_template*.
         :arg complex_kernel: If `True`, prevents vectorization on CPUs.
+        :arg eliminate_empty_output_lists: A Python list of list names
+            for which the empty output lists are eliminated.
 
         *generate_template* may use the following C macros/identifiers:
 
@@ -750,6 +814,11 @@ class ListOfListsBuilder:
         and a second time, for a 'generation' stage where the lists are
         actually filled. A `generate` function that has side effects beyond
         calling `append` is therefore ill-formed.
+
+        .. versionchanged:: 2018.1
+
+            Change *eliminate_empty_output_lists* argument type from `bool` to
+            `list`.
         """
 
         if devices is None:
@@ -767,6 +836,16 @@ class ListOfListsBuilder:
         from pyopencl.tools import parse_arg_list
         self.arg_decls = parse_arg_list(arg_decls)
 
+        # To match with the signature of the user-supplied generate(), arguments
+        # can't appear to have offsets.
+        arg_decls_no_offset = []
+        from pyopencl.tools import VectorArg
+        for arg in self.arg_decls:
+            if isinstance(arg, VectorArg) and arg.with_offset:
+                arg = VectorArg(arg.dtype, arg.name)
+            arg_decls_no_offset.append(arg)
+        self.arg_decls_no_offset = arg_decls_no_offset
+
         self.count_sharing = count_sharing
 
         self.name_prefix = name_prefix
@@ -777,6 +856,20 @@ class ListOfListsBuilder:
 
         self.complex_kernel = complex_kernel
 
+        if eliminate_empty_output_lists is True:
+            eliminate_empty_output_lists = \
+                    [name for name, _ in self.list_names_and_dtypes]
+
+        if eliminate_empty_output_lists is False:
+            eliminate_empty_output_lists = []
+
+        self.eliminate_empty_output_lists = eliminate_empty_output_lists
+        for list_name in self.eliminate_empty_output_lists:
+            if not any(list_name == name for name, _ in self.list_names_and_dtypes):
+                raise ValueError(
+                    "invalid list name '%s' in eliminate_empty_output_lists"
+                    % list_name)
+
     # {{{ kernel generators
 
     @memoize_method
@@ -790,10 +883,36 @@ class ListOfListsBuilder:
                 output_statement="ary[i+1] = item;",
                 devices=self.devices)
 
+    @memoize_method
+    def get_compress_kernel(self, index_dtype):
+        arguments = """
+            __global ${index_t} *count,
+            __global ${index_t} *compressed_counts,
+            __global ${index_t} *nonempty_indices,
+            __global ${index_t} *compressed_indices,
+            __global ${index_t} *num_non_empty_list
+        """
+        arguments = Template(arguments)
+
+        from pyopencl.scan import GenericScanKernel
+        return GenericScanKernel(
+                self.context, index_dtype,
+                arguments=arguments.render(index_t=dtype_to_ctype(index_dtype)),
+                input_expr="count[i] == 0 ? 0 : 1",
+                scan_expr="a+b", neutral="0",
+                output_statement="""
+                    if (i + 1 < N) compressed_indices[i + 1] = item;
+                    if (prev_item != item) {
+                        nonempty_indices[item - 1] = i;
+                        compressed_counts[item - 1] = count[i];
+                    }
+                    if (i + 1 == N) *num_non_empty_list = item;
+                    """,
+                devices=self.devices)
+
     def do_not_vectorize(self):
-        from pytools import any
         return (self.complex_kernel
-                and any(dev.type == cl.device_type.CPU
+                and any(dev.type & cl.device_type.CPU
                     for dev in self.context.devices))
 
     @memoize_method
@@ -802,8 +921,8 @@ class ListOfListsBuilder:
         from pyopencl.tools import VectorArg, OtherArg
         kernel_list_args = [
                 VectorArg(index_dtype, "plb_%s_count" % name)
-                    for name, dtype in self.list_names_and_dtypes
-                    if name not in self.count_sharing]
+                for name, dtype in self.list_names_and_dtypes
+                if name not in self.count_sharing]
 
         user_list_args = []
         for name, dtype in self.list_names_and_dtypes:
@@ -811,7 +930,7 @@ class ListOfListsBuilder:
                 continue
 
             name = "plb_loc_%s_count" % name
-            user_list_args.append(OtherArg("%s *%s" % (
+            user_list_args.append(OtherArg("{} *{}".format(
                 index_ctype, name), name))
 
         kernel_name = self.name_prefix+"_count"
@@ -824,13 +943,16 @@ class ListOfListsBuilder:
                     self.context.devices),
                 debug=self.debug,
                 do_not_vectorize=self.do_not_vectorize(),
+                eliminate_empty_output_lists=self.eliminate_empty_output_lists,
 
                 kernel_list_arg_decl=_get_arg_decl(kernel_list_args),
                 kernel_list_arg_values=_get_arg_list(user_list_args, prefix="&"),
                 user_list_arg_decl=_get_arg_decl(user_list_args),
                 user_list_args=_get_arg_list(user_list_args),
-                user_arg_decl=_get_arg_decl(self.arg_decls),
-                user_args=_get_arg_list(self.arg_decls),
+                user_arg_decl_with_offset=_get_arg_decl(self.arg_decls),
+                user_arg_decl_no_offset=_get_arg_decl(self.arg_decls_no_offset),
+                user_args_no_offset=_get_arg_list(self.arg_decls_no_offset),
+                arg_offset_adjustment=get_arg_offset_adjuster_code(self.arg_decls),
 
                 list_names_and_dtypes=self.list_names_and_dtypes,
                 count_sharing=self.count_sharing,
@@ -874,11 +996,15 @@ class ListOfListsBuilder:
             kernel_list_args.append(
                     VectorArg(index_dtype, "plb_%s_start_index" % name))
 
+            if name in self.eliminate_empty_output_lists:
+                kernel_list_args.append(
+                    VectorArg(index_dtype, "%s_compressed_indices" % name))
+
             index_name = "plb_%s_index" % name
-            user_list_args.append(OtherArg("%s *%s" % (
+            user_list_args.append(OtherArg("{} *{}".format(
                 index_ctype, index_name), index_name))
 
-            kernel_list_arg_values += "%s, &%s, " % (list_name, index_name)
+            kernel_list_arg_values += f"{list_name}, &{index_name}, "
 
         kernel_name = self.name_prefix+"_write"
 
@@ -890,13 +1016,16 @@ class ListOfListsBuilder:
                     self.context.devices),
                 debug=self.debug,
                 do_not_vectorize=self.do_not_vectorize(),
+                eliminate_empty_output_lists=self.eliminate_empty_output_lists,
 
                 kernel_list_arg_decl=_get_arg_decl(kernel_list_args),
                 kernel_list_arg_values=kernel_list_arg_values,
                 user_list_arg_decl=_get_arg_decl(user_list_args),
                 user_list_args=_get_arg_list(user_list_args),
-                user_arg_decl=_get_arg_decl(self.arg_decls),
-                user_args=_get_arg_list(self.arg_decls),
+                user_arg_decl_with_offset=_get_arg_decl(self.arg_decls),
+                user_arg_decl_no_offset=_get_arg_decl(self.arg_decls_no_offset),
+                user_args_no_offset=_get_arg_list(self.arg_decls_no_offset),
+                arg_offset_adjustment=get_arg_offset_adjuster_code(self.arg_decls),
 
                 list_names_and_dtypes=self.list_names_and_dtypes,
                 count_sharing=self.count_sharing,
@@ -925,10 +1054,18 @@ class ListOfListsBuilder:
     def __call__(self, queue, n_objects, *args, **kwargs):
         """
         :arg args: arguments corresponding to arg_decls in the constructor.
-            :class:`pyopencl.array.Array` are not allowed directly and should
-            be passed as their :attr:`pyopencl.array.Array.data` attribute instead.
+            Array-like arguments must be either
+            1D :class:`pyopencl.array.Array` objects or
+            :class:`pyopencl.MemoryObject` objects, of which the latter
+            can be obtained from a :class:`pyopencl.array.Array` using the
+            :attr:`pyopencl.array.Array.data` attribute.
         :arg allocator: optionally, the allocator to use to allocate new
             arrays.
+        :arg omit_lists: An iterable of list names that should *not* be built
+            with this invocation. The kernel code may *not* call ``APPEND_name``
+            for these omitted lists. If it does, undefined behavior will result.
+            The returned *lists* dictionary will not contain an entry for names
+            in *omit_lists*.
         :arg wait_for: |explain-waitfor|
         :returns: a tuple ``(lists, event)``, where
             *lists* a mapping from (built) list names to objects which
@@ -944,7 +1081,22 @@ class ListOfListsBuilder:
 
               This implies that all lists are contiguous.
 
-              *event* is a :class:`pyopencl.Event` for dependency management.
+            If the list name is specified in *eliminate_empty_output_lists*
+            constructor argument, *lists* has two additional attributes
+            ``num_nonempty_lists`` and ``nonempty_indices``
+
+            * ``num_nonempty_lists`` for the number of nonempty lists.
+            * ``nonempty_indices`` for the index of nonempty list in input objects.
+
+            In this case, `starts` has `num_nonempty_lists` + 1 entries. The *i*'s
+            entry is the start of the *i*'th nonempty list, which is generated by
+            the object with index *nonempty_indices[i]*.
+
+            *event* is a :class:`pyopencl.Event` for dependency management.
+
+        .. versionchanged:: 2016.2
+
+            Added omit_lists.
         """
         if n_objects >= int(np.iinfo(np.int32).max):
             index_dtype = np.int64
@@ -953,29 +1105,78 @@ class ListOfListsBuilder:
         index_dtype = np.dtype(index_dtype)
 
         allocator = kwargs.pop("allocator", None)
+        omit_lists = kwargs.pop("omit_lists", [])
         wait_for = kwargs.pop("wait_for", None)
         if kwargs:
             raise TypeError("invalid keyword arguments: '%s'" % ", ".join(kwargs))
 
+        for oml in omit_lists:
+            if not any(oml == name for name, _ in self.list_names_and_dtypes):
+                raise ValueError("invalid list name '%s' in omit_lists")
+
         result = {}
         count_list_args = []
 
+        if wait_for is None:
+            wait_for = []
+        else:
+            # We'll be modifying it below.
+            wait_for = list(wait_for)
+
         count_kernel = self.get_count_kernel(index_dtype)
         write_kernel = self.get_write_kernel(index_dtype)
         scan_kernel = self.get_scan_kernel(index_dtype)
+        if self.eliminate_empty_output_lists:
+            compress_kernel = self.get_compress_kernel(index_dtype)
+
+        data_args = []
+        for i, (arg_descr, arg_val) in enumerate(zip(self.arg_decls, args)):
+            from pyopencl.tools import VectorArg
+            if isinstance(arg_descr, VectorArg):
+                from pyopencl import MemoryObject
+                if isinstance(arg_val, MemoryObject):
+                    data_args.append(arg_val)
+                    if arg_descr.with_offset:
+                        raise ValueError(
+                                "with_offset=True specified for argument %d "
+                                "but the argument is not an array" % i)
+                    continue
+
+                if arg_val.ndim != 1:
+                    raise ValueError("argument %d is a multidimensional array" % i)
+
+                data_args.append(arg_val.base_data)
+                if arg_descr.with_offset:
+                    data_args.append(arg_val.offset)
+                wait_for.extend(arg_val.events)
+            else:
+                data_args.append(arg_val)
+
+        del args
+        data_args = tuple(data_args)
 
         # {{{ allocate memory for counts
 
         for name, dtype in self.list_names_and_dtypes:
             if name in self.count_sharing:
                 continue
+            if name in omit_lists:
+                count_list_args.append(None)
+                continue
 
             counts = cl.array.empty(queue,
                     (n_objects + 1), index_dtype, allocator=allocator)
+            counts[-1] = 0
+            wait_for = wait_for + counts.events
 
             # The scan will turn the "counts" array into the "starts" array
             # in-place.
-            result[name] = BuiltList(starts=counts)
+            if name in self.eliminate_empty_output_lists:
+                result[name] = BuiltList(count=None, starts=counts, lists=None,
+                                         num_nonempty_lists=None,
+                                         nonempty_indices=None)
+            else:
+                result[name] = BuiltList(count=None, starts=counts, lists=None)
             count_list_args.append(counts.data)
 
         # }}}
@@ -983,16 +1184,45 @@ class ListOfListsBuilder:
         if self.debug:
             gsize = (1,)
             lsize = (1,)
-        elif self.complex_kernel and queue.device.type == cl.device_type.CPU:
+        elif self.do_not_vectorize():
             gsize = (4*queue.device.max_compute_units,)
             lsize = (1,)
         else:
-            from pyopencl.array import splay
-            gsize, lsize = splay(queue, n_objects)
+            from pyopencl.array import _splay
+            gsize, lsize = _splay(queue.device, n_objects)
 
         count_event = count_kernel(queue, gsize, lsize,
-                *(tuple(count_list_args) + args + (n_objects,)),
-                **dict(wait_for=wait_for))
+                *(tuple(count_list_args) + data_args + (n_objects,)),
+                wait_for=wait_for)
+
+        compress_events = {}
+        for name, dtype in self.list_names_and_dtypes:
+            if name in omit_lists:
+                continue
+            if name in self.count_sharing:
+                continue
+            if name not in self.eliminate_empty_output_lists:
+                continue
+
+            compressed_counts = cl.array.empty(
+                queue, (n_objects + 1,), index_dtype, allocator=allocator)
+            info_record = result[name]
+            info_record.nonempty_indices = cl.array.empty(
+                queue, (n_objects + 1,), index_dtype, allocator=allocator)
+            info_record.num_nonempty_lists = cl.array.empty(
+                queue, (1,), index_dtype, allocator=allocator)
+            info_record.compressed_indices = cl.array.empty(
+                queue, (n_objects + 1,), index_dtype, allocator=allocator)
+            info_record.compressed_indices[0] = 0
+            compress_events[name] = compress_kernel(
+                info_record.starts,
+                compressed_counts,
+                info_record.nonempty_indices,
+                info_record.compressed_indices,
+                info_record.num_nonempty_lists,
+                wait_for=[count_event] + info_record.compressed_indices.events)
+
+            info_record.starts = compressed_counts
 
         # {{{ run scans
 
@@ -1001,22 +1231,34 @@ class ListOfListsBuilder:
         for name, dtype in self.list_names_and_dtypes:
             if name in self.count_sharing:
                 continue
+            if name in omit_lists:
+                continue
 
             info_record = result[name]
-            starts_ary = info_record.starts
-            evt = scan_kernel(starts_ary, wait_for=[count_event])
+            if name in self.eliminate_empty_output_lists:
+                compress_events[name].wait()
+                num_nonempty_lists = info_record.num_nonempty_lists.get()[0]
+                info_record.num_nonempty_lists = num_nonempty_lists
+                info_record.starts = info_record.starts[:num_nonempty_lists + 1]
+                info_record.nonempty_indices = \
+                    info_record.nonempty_indices[:num_nonempty_lists]
+                info_record.starts[-1] = 0
 
-            # set first entry to zero
-            evt = cl.enqueue_copy(queue, starts_ary.data, index_dtype.type(0),
-                    wait_for=[evt])
-            scan_events.append(evt)
+            starts_ary = info_record.starts
+            if name in self.eliminate_empty_output_lists:
+                evt = scan_kernel(
+                        starts_ary,
+                        size=info_record.num_nonempty_lists,
+                        wait_for=starts_ary.events)
+            else:
+                evt = scan_kernel(starts_ary, wait_for=[count_event],
+                        size=n_objects)
+
+            starts_ary.setitem(0, 0, queue=queue, wait_for=[evt])
+            scan_events.extend(starts_ary.events)
 
             # retrieve count
-            count = np.array(1, index_dtype)
-            cl.enqueue_copy(queue, count, starts_ary.data,
-                    device_offset=index_dtype.itemsize*n_objects)
-
-            info_record.count = int(count)
+            info_record.count = int(starts_ary[-1].get())
 
         # }}}
 
@@ -1024,6 +1266,14 @@ class ListOfListsBuilder:
 
         write_list_args = []
         for name, dtype in self.list_names_and_dtypes:
+            if name in omit_lists:
+                write_list_args.append(None)
+                if name not in self.count_sharing:
+                    write_list_args.append(None)
+                if name in self.eliminate_empty_output_lists:
+                    write_list_args.append(None)
+                continue
+
             if name in self.count_sharing:
                 sharing_from = self.count_sharing[name]
 
@@ -1042,11 +1292,14 @@ class ListOfListsBuilder:
             if name not in self.count_sharing:
                 write_list_args.append(info_record.starts.data)
 
+            if name in self.eliminate_empty_output_lists:
+                write_list_args.append(info_record.compressed_indices.data)
+
         # }}}
 
         evt = write_kernel(queue, gsize, lsize,
-                *(tuple(write_list_args) + args + (n_objects,)),
-                **dict(wait_for=scan_events))
+                *(tuple(write_list_args) + data_args + (n_objects,)),
+                wait_for=scan_events)
 
         return result, evt
 
@@ -1072,7 +1325,7 @@ def _make_cl_int_literal(value, dtype):
     return result
 
 
-class KeyValueSorter(object):
+class KeyValueSorter:
     """Given arrays *values* and *keys* of equal length
     and a number *nkeys* of keys, returns a tuple `(starts,
     lists)`, as follows: *values* and *keys* are sorted
