@@ -1,6 +1,5 @@
 """PyOpenCL compiler cache."""
 
-from __future__ import division
 
 __copyright__ = "Copyright (C) 2011 Andreas Kloeckner"
 
@@ -30,13 +29,12 @@ import sys
 import os
 from pytools import Record
 
-try:
-    import hashlib
-    new_hash = hashlib.md5
-except ImportError:
-    # for Python << 2.5
-    import md5
-    new_hash = md5.new
+import logging
+logger = logging.getLogger(__name__)
+
+
+import hashlib
+new_hash = hashlib.md5
 
 
 def _erase_dir(dir):
@@ -48,7 +46,7 @@ def _erase_dir(dir):
 
 
 def update_checksum(checksum, obj):
-    if isinstance(obj, unicode):
+    if isinstance(obj, str):
         checksum.update(obj.encode("utf8"))
     else:
         checksum.update(obj)
@@ -56,7 +54,7 @@ def update_checksum(checksum, obj):
 
 # {{{ cleanup
 
-class CleanupBase(object):
+class CleanupBase:
     pass
 
 
@@ -100,6 +98,11 @@ class CacheLockManager(CleanupBase):
                     warn("could not obtain cache lock--delete '%s' if necessary"
                             % self.lock_file)
 
+                if attempts > 3 * 60:
+                    raise RuntimeError("waited more than three minutes "
+                            "on the lock file '%s'"
+                            "--something is wrong" % self.lock_file)
+
             cleanup_m.register(self)
 
     def clean_up(self):
@@ -162,8 +165,8 @@ def get_dependencies(src, include_path):
 
                 if included_file_name not in result:
                     try:
-                        src_file = open(included_file_name, "rt")
-                    except IOError:
+                        src_file = open(included_file_name)
+                    except OSError:
                         continue
 
                     try:
@@ -192,7 +195,7 @@ def get_dependencies(src, include_path):
 
     _inner(src)
 
-    result = list((name,) + vals for name, vals in result.iteritems())
+    result = list((name,) + vals for name, vals in result.items())
     result.sort()
 
     return result
@@ -234,10 +237,10 @@ def get_device_cache_id(device):
             device.vendor, device.name, device.version, device.driver_version)
 
 
-def get_cache_key(device, options, src):
+def get_cache_key(device, options_bytes, src):
     checksum = new_hash()
     update_checksum(checksum, src)
-    update_checksum(checksum, " ".join(options))
+    update_checksum(checksum, options_bytes)
     update_checksum(checksum, str(get_device_cache_id(device)))
     return checksum.hexdigest()
 
@@ -265,11 +268,11 @@ def retrieve_from_cache(cache_dir, cache_key):
             # {{{ load info file
 
             try:
-                from cPickle import load
+                from pickle import load
 
                 try:
                     info_file = open(info_path, "rb")
-                except IOError:
+                except OSError:
                     raise _InvalidInfoFile()
 
                 try:
@@ -304,7 +307,7 @@ def retrieve_from_cache(cache_dir, cache_key):
             else:
                 mod_cache_dir_m.reset()
 
-        except:
+        except Exception:
             cleanup_m.error_clean_up()
             raise
     finally:
@@ -317,37 +320,21 @@ class _SourceInfo(Record):
     pass
 
 
-def _create_built_program_from_source_cached(ctx, src, options, devices, cache_dir):
+def _create_built_program_from_source_cached(ctx, src, options_bytes,
+        devices, cache_dir, include_path):
     from os.path import join
 
-    include_path = ["."]
-
-    option_idx = 0
-    while option_idx < len(options):
-        option = options[option_idx].strip()
-        if option.startswith("-I") or option.startswith("/I"):
-            if len(option) == 2:
-                if option_idx+1 < len(options):
-                    include_path.append(options[option_idx+1])
-                option_idx += 2
-            else:
-                include_path.append(option[2:].lstrip())
-                option_idx += 1
-        else:
-            option_idx += 1
-
     if cache_dir is None:
-        from tempfile import gettempdir
-        import getpass
-        cache_dir = join(gettempdir(),
-                "pyopencl-compiler-cache-v2-uid%s-py%s" % (
-                    getpass.getuser(), ".".join(str(i) for i in sys.version_info)))
+        import appdirs
+        cache_dir = join(appdirs.user_cache_dir("pyopencl", "pyopencl"),
+                "pyopencl-compiler-cache-v2-py{}".format(
+                    ".".join(str(i) for i in sys.version_info)))
 
     # {{{ ensure cache directory exists
 
     try:
-        os.mkdir(cache_dir)
-    except OSError, e:
+        os.makedirs(cache_dir)
+    except OSError as e:
         from errno import EEXIST
         if e.errno != EEXIST:
             raise
@@ -357,7 +344,7 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
     if devices is None:
         devices = ctx.devices
 
-    cache_keys = [get_cache_key(device, options, src) for device in devices]
+    cache_keys = [get_cache_key(device, options_bytes, src) for device in devices]
 
     binaries = []
     to_be_built_indices = []
@@ -366,16 +353,20 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
         cache_result = retrieve_from_cache(cache_dir, cache_key)
 
         if cache_result is None:
+            logger.debug("build program: binary cache miss (key: %s)" % cache_key)
+
             to_be_built_indices.append(i)
             binaries.append(None)
             logs.append(None)
         else:
+            logger.debug("build program: binary cache hit (key: %s)" % cache_key)
+
             binary, log = cache_result
             binaries.append(binary)
             logs.append(log)
 
     message = (75*"="+"\n").join(
-            "Build on %s succeeded, but said:\n\n%s" % (dev, log)
+            f"Build on {dev} succeeded, but said:\n\n{log}"
             for dev, log in zip(devices, logs)
             if log is not None and log.strip())
 
@@ -389,6 +380,7 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
 
     result = None
     already_built = False
+    was_cached = not to_be_built_indices
 
     if to_be_built_indices:
         # defeat implementation caches:
@@ -396,8 +388,13 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
         src = src + "\n\n__constant int pyopencl_defeat_cache_%s = 0;" % (
                 uuid4().hex)
 
+        logger.debug("build program: start building program from source on %s"
+                % ", ".join(str(devices[i]) for i in to_be_built_indices))
+
         prg = _cl._Program(ctx, src)
-        prg.build(options, [devices[i] for i in to_be_built_indices])
+        prg.build(options_bytes, [devices[i] for i in to_be_built_indices])
+
+        logger.debug("build program: from-source build complete")
 
         prg_devs = prg.get_info(_cl.program_info.DEVICES)
         prg_bins = prg.get_info(_cl.program_info.BINARIES)
@@ -432,7 +429,6 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
 
                 for i in to_be_built_indices:
                     cache_key = cache_keys[i]
-                    device = devices[i]
                     binary = binaries[i]
 
                     mod_cache_dir_m = ModuleCacheDirManager(cleanup_m,
@@ -449,14 +445,14 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
                     outf.write(binary)
                     outf.close()
 
-                    from cPickle import dump
+                    from pickle import dump
                     info_file = open(info_path, "wb")
                     dump(_SourceInfo(
                         dependencies=get_dependencies(src, include_path),
                         log=logs[i]), info_file)
                     info_file.close()
 
-            except:
+            except Exception:
                 cleanup_m.error_clean_up()
                 raise
         finally:
@@ -464,40 +460,51 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
 
     # }}}
 
-    return result, already_built
+    return result, already_built, was_cached
 
 
-def create_built_program_from_source_cached(ctx, src, options=[], devices=None,
-        cache_dir=None):
+def create_built_program_from_source_cached(ctx, src, options_bytes, devices=None,
+        cache_dir=None, include_path=None):
     try:
+        was_cached = False
+        already_built = False
         if cache_dir is not False:
-            prg, already_built = _create_built_program_from_source_cached(
-                    ctx, src, options, devices, cache_dir)
+            prg, already_built, was_cached = \
+                    _create_built_program_from_source_cached(
+                            ctx, src, options_bytes, devices, cache_dir,
+                            include_path=include_path)
+            if was_cached and not already_built:
+                prg.build(options_bytes, devices)
+                already_built = True
         else:
             prg = _cl._Program(ctx, src)
-            already_built = False
 
-    except Exception, e:
-        raise
+    except Exception as e:
         from pyopencl import Error
-        if (isinstance(e, Error)
-                and e.code == _cl.status_code.BUILD_PROGRAM_FAILURE):
-            # no need to try again
+        build_program_failure = (isinstance(e, Error)
+                and e.code == _cl.status_code.BUILD_PROGRAM_FAILURE)  # noqa pylint:disable=no-member
+
+        # Mac error on intel CPU driver: can't build from cached version.
+        # If we get a build_program_failure from the cached version then
+        # build from source instead, otherwise report the failure.
+        if build_program_failure and not was_cached:
             raise
 
-        from warnings import warn
-        from traceback import format_exc
-        warn("PyOpenCL compiler caching failed with an exception:\n"
-                "[begin exception]\n%s[end exception]"
-                % format_exc())
+        if not build_program_failure:
+            from warnings import warn
+            from traceback import format_exc
+            warn("PyOpenCL compiler caching failed with an exception:\n"
+                    "[begin exception]\n%s[end exception]"
+                    % format_exc())
 
         prg = _cl._Program(ctx, src)
+        was_cached = False
         already_built = False
 
     if not already_built:
-        prg.build(options, devices)
+        prg.build(options_bytes, devices)
 
-    return prg
+    return prg, was_cached
 
 # }}}
 
