@@ -1,5 +1,3 @@
-from __future__ import division
-
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
 __license__ = """
@@ -50,7 +48,7 @@ def has_amd_double_support(dev):
 def reasonable_work_group_size_multiple(dev, ctx=None):
     try:
         return dev.warp_size_nv
-    except:
+    except Exception:
         pass
 
     if ctx is None:
@@ -75,7 +73,7 @@ def nv_compute_capability(dev):
     try:
         return (dev.compute_capability_major_nv,
                 dev.compute_capability_minor_nv)
-    except:
+    except Exception:
         return None
 
 
@@ -211,7 +209,7 @@ def why_not_local_access_conflict_free(dev, itemsize,
     for work_item_axis in range(rank):
 
         bank_accesses = {}
-        for work_item_id in xrange(sim_wi):
+        for work_item_id in range(sim_wi):
             addr = 0
             addr_mult = itemsize
 
@@ -235,13 +233,13 @@ def why_not_local_access_conflict_free(dev, itemsize,
 
             bank = (addr // gran) % bank_count
             bank_accesses.setdefault(bank, []).append(
-                    "w.item %s -> %s" % (work_item_id, idx[::-1]))
+                    "w.item {} -> {}".format(work_item_id, idx[::-1]))
 
         conflict_multiplicity = max(
-                len(acc) for acc in bank_accesses.itervalues())
+                len(acc) for acc in bank_accesses.values())
 
         if conflict_multiplicity > 1:
-            for bank, acc in bank_accesses.iteritems():
+            for bank, acc in bank_accesses.items():
                 if len(acc) == conflict_multiplicity:
                     conflicts.append(
                             (conflict_multiplicity,
@@ -261,8 +259,11 @@ def get_fast_inaccurate_build_options(dev):
     """Return a list of flags valid on device *dev* that enable fast, but
     potentially inaccurate floating point math.
     """
-    return ["-cl-mad-enable", "-cl-fast-relaxed-math",
-        "-cl-no-signed-zeros", "-cl-strict-aliasing"]
+    result = ["-cl-mad-enable", "-cl-fast-relaxed-math",
+        "-cl-no-signed-zeros", ]
+    if dev.vendor.startswith("Advanced Micro") or dev.vendor.startswith("NVIDIA"):
+        result.append("-cl-strict-aliasing")
+    return result
 
 
 def get_simd_group_size(dev, type_size):
@@ -276,16 +277,18 @@ def get_simd_group_size(dev, type_size):
     """
     try:
         return dev.warp_size_nv
-    except:
+    except Exception:
         pass
 
-    lc_vendor = dev.platform.vendor.lower()
-    if "nvidia" in lc_vendor:
+    lc_plat_vendor = dev.platform.vendor.lower()
+    lc_dev_vendor = dev.vendor.lower()
+    if "nvidia" in lc_plat_vendor or "nvidia" in lc_dev_vendor:
         return 32
 
-    if ("advanced micro" in lc_vendor or "ati" in lc_vendor):
+    if ("advanced micro" in lc_plat_vendor or "ati" in lc_plat_vendor
+      or "advanced micro" in lc_dev_vendor or "ati" in lc_dev_vendor):
         if dev.type & cl.device_type.GPU:
-            # Tomasz Rybak says, in response to reduction mishbehaving on the AMD
+            # Tomasz Rybak says, in response to reduction misbehaving on the AMD
             # 'Loveland' APU:
             #
             #    Like in CUDA reduction bug (related to Fermi) it again seems
@@ -310,19 +313,105 @@ def get_simd_group_size(dev, type_size):
 
     if dev.type & cl.device_type.CPU:
         # implicit assumption: Impl. will vectorize
-
-        if type_size == 1:
-            return dev.preferred_vector_width_char
-        elif type_size == 2:
-            return dev.preferred_vector_width_short
-        elif type_size == 4:
-            return dev.preferred_vector_width_float
-        elif type_size == 8:
-            return dev.preferred_vector_width_double
-        else:
-            from warnings import warn
-            warn("unexpected dtype size in get_simd_group on CPU device, "
-                    "guessing group width 1")
-            return 1
+        return 1
 
     return None
+
+
+def get_pocl_version(platform, fallback_value=None):
+    if platform.name != "Portable Computing Language":
+        return None
+
+    import re
+    ver_match = re.match(
+            r"^OpenCL [0-9.]+ pocl ([0-9]+)\.([0-9]+)", platform.version)
+    if ver_match is None:
+        msg = ("pocl version number did not have expected format: '%s'"
+                    % platform.version)
+        if fallback_value is not None:
+            from warnings import warn
+            warn(msg)
+            return fallback_value
+        else:
+            raise ValueError(msg)
+    else:
+        return (int(ver_match.group(1)), int(ver_match.group(2)))
+
+
+_CHECK_FOR_POCL_ARG_COUNT_BUG_CACHE = {}
+
+
+def _check_for_pocl_arg_count_bug(dev, ctx=None):
+    try:
+        return _CHECK_FOR_POCL_ARG_COUNT_BUG_CACHE[dev]
+    except KeyError:
+        pass
+
+    if ctx is None:
+        build_ctx = cl.Context([dev])
+    else:
+        build_ctx = ctx
+
+    prg = cl.Program(build_ctx, """
+            struct two_things
+            {
+                long a;
+                long b;
+            };
+
+            __kernel void test_knl(struct two_things x)
+            {
+            }
+            """).build()
+
+    result = prg.test_knl.num_args == 2
+    _CHECK_FOR_POCL_ARG_COUNT_BUG_CACHE[dev] = result
+
+    return result
+
+
+def has_struct_arg_count_bug(dev, ctx=None):
+    """Checks whether the device is expected to have the
+    `argument counting bug <https://github.com/pocl/pocl/issues/197>`_.
+    """
+
+    if dev.platform.name == "Apple" and dev.type & cl.device_type.CPU:
+        return "apple"
+    if dev.platform.name == "Portable Computing Language":
+        pocl_version = get_pocl_version(dev.platform, fallback_value=(0.14))
+        if pocl_version <= (0, 13):
+            return "pocl"
+        elif pocl_version <= (0, 14) and _check_for_pocl_arg_count_bug(dev, ctx):
+            return "pocl"
+
+    return False
+
+
+def _may_have_svm(dev):
+    has_svm = (dev.platform._get_cl_version() >= (2, 0)
+            and cl.get_cl_header_version() >= (2, 0))
+
+    if dev.platform.name == "Portable Computing Language":
+        has_svm = (
+                get_pocl_version(dev.platform) >= (1, 0)
+                and cl.get_cl_header_version() >= (2, 0))
+
+    return has_svm
+
+
+def has_coarse_grain_buffer_svm(dev):
+    return (_may_have_svm(dev)
+            and bool(dev.svm_capabilities
+                & cl.device_svm_capabilities.COARSE_GRAIN_BUFFER))
+
+
+def has_fine_grain_buffer_svm(dev):
+    return (_may_have_svm(dev)
+            and bool(dev.svm_capabilities
+                & cl.device_svm_capabilities.FINE_GRAIN_BUFFER))
+
+
+def has_fine_grain_system_svm(dev):
+    return (_may_have_svm(dev)
+            and bool(dev.svm_capabilities
+                & cl.device_svm_capabilities.FINE_GRAIN_SYSTEM))
