@@ -1,6 +1,5 @@
 """Scan primitive."""
 
-from __future__ import division
 
 __copyright__ = """
 Copyright 2011-2012 Andreas Kloeckner
@@ -20,14 +19,9 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
-Derived from thrust/detail/backend/cuda/detail/fast_scan.inl
-within the Thrust project, https://code.google.com/p/thrust/
-
+Derived from code within the Thrust project, https://github.com/thrust/thrust/
 """
 
-# Direct link to thrust source:
-# https://code.google.com/p/thrust/source/browse/thrust/detail/backend/cuda/detail/fast_scan.inl # noqa
-
 import numpy as np
 
 import pyopencl as cl
@@ -35,11 +29,19 @@ import pyopencl.array  # noqa
 from pyopencl.tools import (dtype_to_ctype, bitlog2,
         KernelTemplateBase, _process_code_for_macro,
         get_arg_list_scalar_arg_dtypes,
-        context_dependent_memoize)
+        context_dependent_memoize,
+        _NumpyTypesKeyBuilder,
+        get_arg_offset_adjuster_code)
 
 import pyopencl._mymako as mako
 from pyopencl._cluda import CLUDA_PREAMBLE
 
+from pytools.persistent_dict import WriteOncePersistentDict
+
+
+import logging
+logger = logging.getLogger(__name__)
+
 
 # {{{ preamble
 
@@ -125,7 +127,7 @@ SCAN_INTERVALS_SOURCE = SHARED_PREAMBLE + r"""//CL//
 
 KERNEL
 REQD_WG_SIZE(WG_SIZE, 1, 1)
-void ${name_prefix}_scan_intervals(
+void ${kernel_name}(
     ${argument_signature},
     GLOBAL_MEM scan_type *restrict partial_scan_buffer,
     const index_type N,
@@ -142,6 +144,8 @@ void ${name_prefix}_scan_intervals(
     %endif
     )
 {
+    ${arg_offset_adjustment}
+
     // index K in first dimension used for carry storage
     %if use_bank_conflict_avoidance:
         // Avoid bank conflicts by adding a single 32-bit value to the size of
@@ -151,16 +155,14 @@ void ${name_prefix}_scan_intervals(
             scan_type value;
             int dummy;
         };
-        LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE + 1];
     %else:
         struct wrapped_scan_type
         {
             scan_type value;
         };
-
-        // padded in WG_SIZE to avoid bank conflicts
-        LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE];
     %endif
+    // padded in WG_SIZE to avoid bank conflicts
+    LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE + 1];
 
     %if is_segmented:
         LOCAL_MEM char l_segment_start_flags[K][WG_SIZE];
@@ -296,8 +298,11 @@ void ${name_prefix}_scan_intervals(
 
             if (LID_0 == 0 && unit_base != interval_begin)
             {
+                scan_type tmp = ldata[K][WG_SIZE - 1].value;
+                scan_type tmp_aux = ldata[0][0].value;
+
                 ldata[0][0].value = SCAN_EXPR(
-                    ldata[K][WG_SIZE - 1].value, ldata[0][0].value,
+                    tmp, tmp_aux,
                     %if is_segmented:
                         (l_segment_start_flags[0][0])
                     %else:
@@ -320,16 +325,17 @@ void ${name_prefix}_scan_intervals(
                 const index_type offset_end = interval_end - unit_base;
             %endif
 
-            for(index_type k = 1; k < K; k++)
+            for (index_type k = 1; k < K; k++)
             {
                 %if is_tail:
-                if (K * LID_0 + k < offset_end)
+                if ((index_type) (K * LID_0 + k) < offset_end)
                 %endif
                 {
                     scan_type tmp = ldata[k][LID_0].value;
-                    index_type seq_i = unit_base + K*LID_0 + k;
 
                     %if is_segmented:
+                    index_type seq_i = unit_base + K*LID_0 + k;
+
                     if (l_segment_start_flags[k][LID_0])
                     {
                         first_segment_start_in_k_group = min(
@@ -611,6 +617,8 @@ void ${name_prefix}_final_update(
     %endif
     )
 {
+    ${arg_offset_adjustment}
+
     %if use_lookbehind_update:
         LOCAL_MEM scan_type ldata[WG_SIZE];
     %endif
@@ -742,6 +750,7 @@ def _round_down_to_power_of_2(val):
     assert result <= val
     return result
 
+
 _PREFIX_WORDS = set("""
         ldata partial_scan_buffer global scan_offset
         segment_start_in_k_group carry
@@ -760,7 +769,7 @@ _PREFIX_WORDS = set("""
         group_base seg_end my_val DEBUG ARGS
         ints_to_store ints_per_wg scan_types_per_int linear_index
         linear_scan_data_idx dest src store_base wrapped_scan_type
-        dummy
+        dummy scan_tmp tmp_aux
 
         LID_2 LID_1 LID_0
         LDIM_0 LDIM_1 LDIM_2
@@ -783,8 +792,9 @@ _IGNORED_WORDS = set("""
         pragma __attribute__ __global __kernel __local
         get_local_size get_local_id cl_khr_fp64 reqd_work_group_size
         get_num_groups barrier get_group_id
+        CL_VERSION_1_1 __OPENCL_C_VERSION__ 120
 
-        _final_update _scan_intervals _debug_scan
+        _final_update _debug_scan kernel_name
 
         positions all padded integer its previous write based writes 0
         has local worth scan_expr to read cannot not X items False bank
@@ -821,6 +831,8 @@ _IGNORED_WORDS = set("""
 
         a b prev_item i last_item prev_value
         N NO_SEG_BOUNDARY across_seg_boundary
+
+        arg_offset_adjustment
         """.split())
 
 
@@ -849,11 +861,72 @@ def _make_template(s):
 
     return mako.template.Template(s, strict_undefined=True)
 
-from pytools import Record
+
+from pytools import Record, RecordWithoutPickling
 
 
-class _ScanKernelInfo(Record):
-    pass
+class _GeneratedScanKernelInfo(Record):
+
+    __slots__ = [
+            "scan_src",
+            "kernel_name",
+            "scalar_arg_dtypes",
+            "wg_size",
+            "k_group_size"]
+
+    def __init__(self, scan_src, kernel_name, scalar_arg_dtypes, wg_size,
+            k_group_size):
+        Record.__init__(self,
+                scan_src=scan_src,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
+
+    def build(self, context, options):
+        program = cl.Program(context, self.scan_src).build(options)
+        kernel = getattr(program, self.kernel_name)
+        kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
+        return _BuiltScanKernelInfo(
+                kernel=kernel,
+                wg_size=self.wg_size,
+                k_group_size=self.k_group_size)
+
+
+class _BuiltScanKernelInfo(RecordWithoutPickling):
+
+    __slots__ = ["kernel", "wg_size", "k_group_size"]
+
+    def __init__(self, kernel, wg_size, k_group_size):
+        RecordWithoutPickling.__init__(self,
+                kernel=kernel,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
+
+
+class _GeneratedFinalUpdateKernelInfo(Record):
+
+    def __init__(self, source, kernel_name, scalar_arg_dtypes, update_wg_size):
+        Record.__init__(self,
+                source=source,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                update_wg_size=update_wg_size)
+
+    def build(self, context, options):
+        program = cl.Program(context, self.source).build(options)
+        kernel = getattr(program, self.kernel_name)
+        kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
+        return _BuiltFinalUpdateKernelInfo(kernel, self.update_wg_size)
+
+
+class _BuiltFinalUpdateKernelInfo(RecordWithoutPickling):
+    __slots__ = ["kernel", "update_wg_size"]
+
+    def __init__(self, kernel, update_wg_size):
+        RecordWithoutPickling.__init__(self,
+                kernel=kernel,
+                update_wg_size=update_wg_size)
 
 # }}}
 
@@ -862,7 +935,7 @@ class ScanPerformanceWarning(UserWarning):
     pass
 
 
-class _GenericScanKernelBase(object):
+class _GenericScanKernelBase:
     # {{{ constructor, argument processing
 
     def __init__(self, ctx, dtype,
@@ -928,7 +1001,7 @@ class _GenericScanKernelBase(object):
             resulting in a C `bool` value that determines whether a new
             scan segments starts at index *i*.  If given, makes the scan a
             segmented scan. Has access to the current index `i`, the result
-            of *input_expr* as a, and in addition may use *arguments* and
+            of *input_expr* as `a`, and in addition may use *arguments* and
             *input_fetch_expr* variables just like *input_expr*.
 
             If it returns true, then previous sums will not spill over into the
@@ -1006,7 +1079,6 @@ class _GenericScanKernelBase(object):
 
         # {{{ set up shared code dict
 
-        from pytools import all
         from pyopencl.characterize import has_double_support
 
         self.code_variables = dict(
@@ -1026,12 +1098,47 @@ class _GenericScanKernelBase(object):
                 has_double_support(dev) for dev in devices),
             )
 
+        index_typename = dtype_to_ctype(self.index_dtype)
+        scan_typename = dtype_to_ctype(dtype)
+
+        # This key is meant to uniquely identify the non-device parameters for
+        # the scan kernel.
+        self.kernel_key = (
+            self.dtype,
+            tuple(arg.declarator() for arg in self.parsed_args),
+            self.input_expr,
+            scan_expr,
+            neutral,
+            output_statement,
+            is_segment_start_expr,
+            tuple(input_fetch_exprs),
+            index_dtype,
+            name_prefix,
+            preamble,
+            # These depend on dtype_to_ctype(), so their value is independent of
+            # the other variables.
+            index_typename,
+            scan_typename,
+            )
+
         # }}}
 
+        self.use_lookbehind_update = "prev_item" in self.output_statement
+        self.store_segment_start_flags = (
+                self.is_segmented and self.use_lookbehind_update)
+
         self.finish_setup()
 
     # }}}
 
+    def finish_setup(self):
+        raise NotImplementedError
+
+
+generic_scan_kernel_cache = WriteOncePersistentDict(
+        "pyopencl-generated-scan-kernel-cache-v1",
+        key_builder=_NumpyTypesKeyBuilder())
+
 
 class GenericScanKernel(_GenericScanKernelBase):
     """Generates and executes code that performs prefix sums ("scans") on
@@ -1048,14 +1155,57 @@ class GenericScanKernel(_GenericScanKernelBase):
                 output_statement="ary[i+1] = item;")
 
         a = cl.array.arange(queue, 10000, dtype=np.int32)
-        scan_kernel(a, queue=queue)
+        knl(a, queue=queue)
 
     """
 
     def finish_setup(self):
-        use_lookbehind_update = "prev_item" in self.output_statement
-        self.store_segment_start_flags = self.is_segmented and use_lookbehind_update
+        # Before generating the kernel, see if it's cached.
+        from pyopencl.cache import get_device_cache_id
+        devices_key = tuple(get_device_cache_id(device)
+                for device in self.devices)
 
+        cache_key = (self.kernel_key, devices_key)
+
+        from_cache = False
+
+        try:
+            result = generic_scan_kernel_cache[cache_key]
+            from_cache = True
+            logger.debug(
+                    "cache hit for generated scan kernel '%s'" % self.name_prefix)
+            (
+                self.first_level_scan_gen_info,
+                self.second_level_scan_gen_info,
+                self.final_update_gen_info) = result
+        except KeyError:
+            pass
+
+        if not from_cache:
+            logger.debug(
+                    "cache miss for generated scan kernel '%s'" % self.name_prefix)
+            self._finish_setup_impl()
+
+            result = (self.first_level_scan_gen_info,
+                      self.second_level_scan_gen_info,
+                      self.final_update_gen_info)
+
+            generic_scan_kernel_cache.store_if_not_present(cache_key, result)
+
+        # Build the kernels.
+        self.first_level_scan_info = self.first_level_scan_gen_info.build(
+                self.context, self.options)
+        del self.first_level_scan_gen_info
+
+        self.second_level_scan_info = self.second_level_scan_gen_info.build(
+                self.context, self.options)
+        del self.second_level_scan_gen_info
+
+        self.final_update_info = self.final_update_gen_info.build(
+                self.context, self.options)
+        del self.final_update_gen_info
+
+    def _finish_setup_impl(self):
         # {{{ find usable workgroup/k-group size, build first-level scan
 
         trip_count = 0
@@ -1064,6 +1214,10 @@ class GenericScanKernel(_GenericScanKernelBase):
                 dev.local_mem_size
                 for dev in self.devices)
 
+        if "CUDA" in self.devices[0].platform.name:
+            # not sure where these go, but roughly this much seems unavailable.
+            avail_local_mem -= 0x400
+
         is_cpu = self.devices[0].type & cl.device_type.CPU
         is_gpu = self.devices[0].type & cl.device_type.GPU
 
@@ -1076,8 +1230,17 @@ class GenericScanKernel(_GenericScanKernelBase):
             max_scan_wg_size = min(dev.max_work_group_size for dev in self.devices)
             wg_size_multiples = 64
 
+        # Intel beignet fails "Out of shared local memory" in test_scan int64
+        # and asserts in test_sort with this enabled:
+        # https://github.com/inducer/pyopencl/pull/238
+        # A beignet bug report (outside of pyopencl) suggests packed structs
+        # (which this is) can even give wrong results:
+        # https://bugs.freedesktop.org/show_bug.cgi?id=98717
+        # TODO: does this also affect Intel Compute Runtime?
         use_bank_conflict_avoidance = (
-                self.dtype.itemsize > 4 and self.dtype.itemsize % 8 == 0 and is_gpu)
+                self.dtype.itemsize > 4 and self.dtype.itemsize % 8 == 0
+                and is_gpu
+                and "beignet" not in self.devices[0].platform.version.lower())
 
         # k_group_size should be a power of two because of in-kernel
         # division by that number.
@@ -1090,26 +1253,25 @@ class GenericScanKernel(_GenericScanKernelBase):
                 k_group_size = 2**k_exp
                 lmem_use = self.get_local_mem_use(wg_size, k_group_size,
                         use_bank_conflict_avoidance)
-                if lmem_use + 256 <= avail_local_mem:
+                if lmem_use <= avail_local_mem:
                     solutions.append((wg_size*k_group_size, k_group_size, wg_size))
 
         if is_gpu:
-            from pytools import any
             for wg_size_floor in [256, 192, 128]:
                 have_sol_above_floor = any(wg_size >= wg_size_floor
                         for _, _, wg_size in solutions)
 
                 if have_sol_above_floor:
                     # delete all solutions not meeting the wg size floor
-                    solutions = [(total, k_group_size, wg_size)
-                            for total, k_group_size, wg_size in solutions
-                            if wg_size >= wg_size_floor]
+                    solutions = [(total, try_k_group_size, try_wg_size)
+                            for total, try_k_group_size, try_wg_size in solutions
+                            if try_wg_size >= wg_size_floor]
                     break
 
         _, k_group_size, max_scan_wg_size = max(solutions)
 
         while True:
-            candidate_scan_info = self.build_scan_kernel(
+            candidate_scan_gen_info = self.generate_scan_kernel(
                     max_scan_wg_size, self.parsed_args,
                     _process_code_for_macro(self.input_expr),
                     self.is_segment_start_expr,
@@ -1119,6 +1281,9 @@ class GenericScanKernel(_GenericScanKernelBase):
                     k_group_size=k_group_size,
                     use_bank_conflict_avoidance=use_bank_conflict_avoidance)
 
+            candidate_scan_info = candidate_scan_gen_info.build(
+                    self.context, self.options)
+
             # Will this device actually let us execute this kernel
             # at the desired work group size? Building it is the
             # only way to find out.
@@ -1136,7 +1301,7 @@ class GenericScanKernel(_GenericScanKernelBase):
             trip_count += 1
             assert trip_count <= 20
 
-        self.first_level_scan_info = candidate_scan_info
+        self.first_level_scan_gen_info = candidate_scan_gen_info
         assert (_round_down_to_power_of_2(candidate_scan_info.wg_size)
                 == candidate_scan_info.wg_size)
 
@@ -1164,7 +1329,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         else:
             second_level_build_kwargs["is_segment_start_expr"] = None
 
-        self.second_level_scan_info = self.build_scan_kernel(
+        self.second_level_scan_gen_info = self.generate_scan_kernel(
                 max_scan_wg_size,
                 arguments=second_level_arguments,
                 input_expr="interval_sums[i]",
@@ -1177,26 +1342,22 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         # }}}
 
-        # {{{ build final update kernel
+        # {{{ generate final update kernel
 
-        self.update_wg_size = min(max_scan_wg_size, 256)
+        update_wg_size = min(max_scan_wg_size, 256)
 
         final_update_tpl = _make_template(UPDATE_SOURCE)
         final_update_src = str(final_update_tpl.render(
-            wg_size=self.update_wg_size,
+            wg_size=update_wg_size,
             output_statement=self.output_statement,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(self.parsed_args),
             argument_signature=", ".join(
                 arg.declarator() for arg in self.parsed_args),
             is_segment_start_expr=self.is_segment_start_expr,
             input_expr=_process_code_for_macro(self.input_expr),
-            use_lookbehind_update=use_lookbehind_update,
+            use_lookbehind_update=self.use_lookbehind_update,
             **self.code_variables))
 
-        final_update_prg = cl.Program(
-                self.context, final_update_src).build(self.options)
-        self.final_update_knl = getattr(
-                final_update_prg,
-                self.name_prefix+"_final_update")
         update_scalar_arg_dtypes = (
                 get_arg_list_scalar_arg_dtypes(self.parsed_args)
                 + [self.index_dtype, self.index_dtype, None, None])
@@ -1205,7 +1366,12 @@ class GenericScanKernel(_GenericScanKernelBase):
             update_scalar_arg_dtypes.append(None)
         if self.store_segment_start_flags:
             update_scalar_arg_dtypes.append(None)  # g_segment_start_flags
-        self.final_update_knl.set_scalar_arg_dtypes(update_scalar_arg_dtypes)
+
+        self.final_update_gen_info = _GeneratedFinalUpdateKernelInfo(
+                final_update_src,
+                self.name_prefix + "_final_update",
+                update_scalar_arg_dtypes,
+                update_wg_size)
 
         # }}}
 
@@ -1236,10 +1402,10 @@ class GenericScanKernel(_GenericScanKernelBase):
 
                 + k_group_size*wg_size*sum(
                     arg_dtypes[arg_name].itemsize
-                    for arg_name, ife_offsets in fetch_expr_offsets.items()
+                    for arg_name, ife_offsets in list(fetch_expr_offsets.items())
                     if -1 in ife_offsets or len(ife_offsets) > 1))
 
-    def build_scan_kernel(self, max_wg_size, arguments, input_expr,
+    def generate_scan_kernel(self, max_wg_size, arguments, input_expr,
             is_segment_start_expr, input_fetch_exprs, is_first_level,
             store_segment_start_flags, k_group_size,
             use_bank_conflict_avoidance):
@@ -1249,37 +1415,42 @@ class GenericScanKernel(_GenericScanKernelBase):
         wg_size = _round_down_to_power_of_2(
                 min(max_wg_size, 256))
 
+        kernel_name = self.code_variables["name_prefix"]
+        if is_first_level:
+            kernel_name += "_lev1"
+        else:
+            kernel_name += "_lev2"
+
         scan_tpl = _make_template(SCAN_INTERVALS_SOURCE)
         scan_src = str(scan_tpl.render(
             wg_size=wg_size,
             input_expr=input_expr,
             k_group_size=k_group_size,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(arguments),
             argument_signature=", ".join(arg.declarator() for arg in arguments),
             is_segment_start_expr=is_segment_start_expr,
             input_fetch_exprs=input_fetch_exprs,
             is_first_level=is_first_level,
             store_segment_start_flags=store_segment_start_flags,
             use_bank_conflict_avoidance=use_bank_conflict_avoidance,
+            kernel_name=kernel_name,
             **self.code_variables))
 
-        prg = cl.Program(self.context, scan_src).build(self.options)
-
-        knl = getattr(
-                prg,
-                self.code_variables["name_prefix"]+"_scan_intervals")
-
         scalar_arg_dtypes.extend(
-                (None, self.index_dtype, self. index_dtype))
+                (None, self.index_dtype, self.index_dtype))
         if is_first_level:
             scalar_arg_dtypes.append(None)  # interval_results
         if self.is_segmented and is_first_level:
             scalar_arg_dtypes.append(None)  # g_first_segment_start_in_interval
         if store_segment_start_flags:
             scalar_arg_dtypes.append(None)  # g_segment_start_flags
-        knl.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
-        return _ScanKernelInfo(
-                kernel=knl, wg_size=wg_size, knl=knl, k_group_size=k_group_size)
+        return _GeneratedScanKernelInfo(
+                scan_src=scan_src,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
 
     # }}}
 
@@ -1291,6 +1462,11 @@ class GenericScanKernel(_GenericScanKernelBase):
         n = kwargs.get("size")
         wait_for = kwargs.get("wait_for")
 
+        if wait_for is None:
+            wait_for = []
+        else:
+            wait_for = list(wait_for)
+
         if len(args) != len(self.parsed_args):
             raise TypeError("expected %d arguments, got %d" %
                     (len(self.parsed_args), len(args)))
@@ -1307,10 +1483,13 @@ class GenericScanKernel(_GenericScanKernelBase):
             return cl.enqueue_marker(queue, wait_for=wait_for)
 
         data_args = []
-        from pyopencl.tools import VectorArg
         for arg_descr, arg_val in zip(self.parsed_args, args):
+            from pyopencl.tools import VectorArg
             if isinstance(arg_descr, VectorArg):
-                data_args.append(arg_val.data)
+                data_args.append(arg_val.base_data)
+                if arg_descr.with_offset:
+                    data_args.append(arg_val.offset)
+                wait_for.extend(arg_val.events)
             else:
                 data_args.append(arg_val)
 
@@ -1361,7 +1540,7 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         l1_evt = l1_info.kernel(
                 queue, (num_intervals,), (l1_info.wg_size,),
-                *scan1_args, **dict(g_times_l=True, wait_for=wait_for))
+                *scan1_args, g_times_l=True, wait_for=wait_for)
 
         # }}}
 
@@ -1381,7 +1560,7 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         l2_evt = l2_info.kernel(
                 queue, (1,), (l1_info.wg_size,),
-                *scan2_args, **dict(g_times_l=True, wait_for=[l1_evt]))
+                *scan2_args, g_times_l=True, wait_for=[l1_evt])
 
         # }}}
 
@@ -1394,14 +1573,16 @@ class GenericScanKernel(_GenericScanKernelBase):
         if self.store_segment_start_flags:
             upd_args.append(segment_start_flags.data)
 
-        return self.final_update_knl(
-                queue, (num_intervals,), (self.update_wg_size,),
-                *upd_args, **dict(g_times_l=True, wait_for=[l2_evt]))
+        return self.final_update_info.kernel(
+                queue, (num_intervals,),
+                (self.final_update_info.update_wg_size,),
+                *upd_args, g_times_l=True, wait_for=[l2_evt])
 
         # }}}
 
 # }}}
 
+
 # {{{ debug kernel
 
 DEBUG_SCAN_TEMPLATE = SHARED_PREAMBLE + r"""//CL//
@@ -1409,11 +1590,14 @@ DEBUG_SCAN_TEMPLATE = SHARED_PREAMBLE + r"""//CL//
 KERNEL
 REQD_WG_SIZE(1, 1, 1)
 void ${name_prefix}_debug_scan(
+    __global scan_type *scan_tmp,
     ${argument_signature},
     const index_type N)
 {
-    scan_type item = ${neutral};
-    scan_type prev_item;
+    scan_type current = ${neutral};
+    scan_type prev;
+
+    ${arg_offset_adjustment}
 
     for (index_type i = 0; i < N; ++i)
     {
@@ -1421,7 +1605,7 @@ void ${name_prefix}_debug_scan(
             ${arg_ctypes[arg_name]} ${name};
             %if ife_offset < 0:
                 if (i+${ife_offset} >= 0)
-                    ${name} = ${arg_name}[i+offset];
+                    ${name} = ${arg_name}[i+${ife_offset}];
             %else:
                 ${name} = ${arg_name}[i];
             %endif
@@ -1429,18 +1613,31 @@ void ${name_prefix}_debug_scan(
 
         scan_type my_val = INPUT_EXPR(i);
 
-        prev_item = item;
+        prev = current;
         %if is_segmented:
             bool is_seg_start = IS_SEG_START(i, my_val);
         %endif
 
-        item = SCAN_EXPR(prev_item, my_val,
+        current = SCAN_EXPR(prev, my_val,
             %if is_segmented:
                 is_seg_start
             %else:
                 false
             %endif
             );
+        scan_tmp[i] = current;
+    }
+
+    scan_type last_item = scan_tmp[N-1];
+
+    for (index_type i = 0; i < N; ++i)
+    {
+        scan_type item = scan_tmp[i];
+        scan_type prev_item;
+        if (i)
+            prev_item = scan_tmp[i-1];
+        else
+            prev_item = ${neutral};
 
         {
             ${output_statement};
@@ -1455,6 +1652,7 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         scan_tpl = _make_template(DEBUG_SCAN_TEMPLATE)
         scan_src = str(scan_tpl.render(
             output_statement=self.output_statement,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(self.parsed_args),
             argument_signature=", ".join(
                 arg.declarator() for arg in self.parsed_args),
             is_segment_start_expr=self.is_segment_start_expr,
@@ -1467,7 +1665,8 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         self.kernel = getattr(
                 scan_prg, self.name_prefix+"_debug_scan")
         scalar_arg_dtypes = (
-                get_arg_list_scalar_arg_dtypes(self.parsed_args)
+                [None]
+                + get_arg_list_scalar_arg_dtypes(self.parsed_args)
                 + [self.index_dtype])
         self.kernel.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
@@ -1479,6 +1678,12 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         n = kwargs.get("size")
         wait_for = kwargs.get("wait_for")
 
+        if wait_for is None:
+            wait_for = []
+        else:
+            # We'll be modifying it below.
+            wait_for = list(wait_for)
+
         if len(args) != len(self.parsed_args):
             raise TypeError("expected %d arguments, got %d" %
                     (len(self.parsed_args), len(args)))
@@ -1490,18 +1695,25 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         if n is None:
             n, = first_array.shape
 
-        data_args = []
+        scan_tmp = cl.array.empty(queue,
+                n, dtype=self.dtype,
+                allocator=allocator)
+
+        data_args = [scan_tmp.data]
         from pyopencl.tools import VectorArg
         for arg_descr, arg_val in zip(self.parsed_args, args):
             if isinstance(arg_descr, VectorArg):
-                data_args.append(arg_val.data)
+                data_args.append(arg_val.base_data)
+                if arg_descr.with_offset:
+                    data_args.append(arg_val.offset)
+                wait_for.extend(arg_val.events)
             else:
                 data_args.append(arg_val)
 
         # }}}
 
         return self.kernel(queue, (1,), (1,),
-                *(data_args + [n]), **dict(wait_for=wait_for))
+                *(data_args + [n]), wait_for=wait_for)
 
 # }}}
 
@@ -1515,7 +1727,7 @@ class _LegacyScanKernelBase(GenericScanKernel):
         scan_ctype = dtype_to_ctype(dtype)
         GenericScanKernel.__init__(self,
                 ctx, dtype,
-                arguments="__global %s *input_ary, __global %s *output_ary" % (
+                arguments="__global {} *input_ary, __global {} *output_ary".format(
                     scan_ctype, scan_ctype),
                 input_expr="input_ary[i]",
                 scan_expr=scan_expr,
@@ -1523,6 +1735,10 @@ class _LegacyScanKernelBase(GenericScanKernel):
                 output_statement=self.ary_output_statement,
                 options=options, preamble=preamble, devices=devices)
 
+    @property
+    def ary_output_statement(self):
+        raise NotImplementedError
+
     def __call__(self, input_ary, output_ary=None, allocator=None, queue=None):
         allocator = allocator or input_ary.allocator
         queue = queue or input_ary.queue or output_ary.queue
@@ -1530,7 +1746,7 @@ class _LegacyScanKernelBase(GenericScanKernel):
         if output_ary is None:
             output_ary = input_ary
 
-        if isinstance(output_ary, (str, unicode)) and output_ary == "new":
+        if isinstance(output_ary, (str, str)) and output_ary == "new":
             output_ary = cl.array.empty_like(input_ary, allocator=allocator)
 
         if input_ary.shape != output_ary.shape:
