@@ -1,6 +1,5 @@
 #! /usr/bin/env python
 
-from __future__ import division, with_statement
 
 __copyright__ = "Copyright (C) 2013 Andreas Kloeckner"
 
@@ -24,6 +23,9 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
+# avoid spurious: pytest.mark.parametrize is not callable
+# pylint: disable=not-callable
+
 import numpy as np
 import numpy.linalg as la
 import sys
@@ -36,8 +38,10 @@ import pyopencl as cl
 import pyopencl.array as cl_array  # noqa
 from pyopencl.tools import (  # noqa
         pytest_generate_tests_for_pyopencl as pytest_generate_tests)
-from pyopencl.characterize import has_double_support
-from pyopencl.scan import InclusiveScanKernel, ExclusiveScanKernel
+from pyopencl.characterize import has_double_support, has_struct_arg_count_bug
+from pyopencl.scan import (InclusiveScanKernel, ExclusiveScanKernel,
+        GenericScanKernel, GenericDebugScanKernel)
+from pyopencl.characterize import get_pocl_version
 
 
 # {{{ elementwise
@@ -72,7 +76,7 @@ def test_elwise_kernel_with_options(ctx_factory):
 
     in_gpu = clrand(queue, (50,), np.float32)
 
-    options = ['-D', 'ADD_ONE']
+    options = ["-D", "ADD_ONE"]
     add_one = ElementwiseKernel(
         context,
         "float* out, const float *in",
@@ -156,9 +160,9 @@ def test_if_positive(ctx_factory):
 
     from pyopencl.clrandom import rand as clrand
 
-    l = 20000
-    a_gpu = clrand(queue, (l,), np.float32)
-    b_gpu = clrand(queue, (l,), np.float32)
+    ary_len = 20000
+    a_gpu = clrand(queue, (ary_len,), np.float32)
+    b_gpu = clrand(queue, (ary_len,), np.float32)
     a = a_gpu.get()
     b = b_gpu.get()
 
@@ -242,11 +246,43 @@ def test_sum(ctx_factory):
                 slice(1000, 3000),
                 slice(1000, -3000),
                 slice(1000, None),
+                slice(1000, None, 3),
+                slice(1000, 1000),
                 ]:
             sum_a = np.sum(a[slc])
-            sum_a_gpu = cl_array.sum(a_gpu[slc]).get()
 
-            assert abs(sum_a_gpu - sum_a) / abs(sum_a) < 1e-4
+            if sum_a:
+                ref_divisor = abs(sum_a)
+            else:
+                ref_divisor = 1
+
+            if slc.step is None:
+                sum_a_gpu = cl_array.sum(a_gpu[slc]).get()
+                assert abs(sum_a_gpu - sum_a) / ref_divisor < 1e-4
+
+            sum_a_gpu_2 = cl_array.sum(a_gpu, slice=slc).get()
+            assert abs(sum_a_gpu_2 - sum_a) / ref_divisor < 1e-4
+
+
+def test_sum_without_data(ctx_factory):
+    from pytest import importorskip
+    importorskip("mako")
+
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    n = 2000
+
+    from pyopencl.reduction import ReductionKernel
+    red = ReductionKernel(context, np.int32,
+            neutral="0",
+            reduce_expr="a+b", map_expr="i",
+            arguments=[])
+
+    result_dev = red(range=slice(n), queue=queue).get()
+    result_ref = n*(n-1)//2
+
+    assert result_dev == result_ref
 
 
 def test_minmax(ctx_factory):
@@ -323,9 +359,14 @@ def test_dot(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
+    dev = context.devices[0]
+
     dtypes = [np.float32, np.complex64]
-    if has_double_support(context.devices[0]):
-        dtypes.extend([np.float64, np.complex128])
+    if has_double_support(dev):
+        if has_struct_arg_count_bug(dev) == "apple":
+            dtypes.extend([np.float64])
+        else:
+            dtypes.extend([np.float64, np.complex128])
 
     for a_dtype in dtypes:
         for b_dtype in dtypes:
@@ -340,10 +381,21 @@ def test_dot(ctx_factory):
 
             assert abs(dot_ab_gpu - dot_ab) / abs(dot_ab) < 1e-4
 
-            vdot_ab = np.vdot(a, b)
+            try:
+                vdot_ab = np.vdot(a, b)
+            except NotImplementedError:
+                import sys
+                is_pypy = "__pypy__" in sys.builtin_module_names
+                if is_pypy:
+                    print("PYPY: VDOT UNIMPLEMENTED")
+                    continue
+                else:
+                    raise
+
             vdot_ab_gpu = cl_array.vdot(a_gpu, b_gpu).get()
 
-            assert abs(vdot_ab_gpu - vdot_ab) / abs(vdot_ab) < 1e-4
+            rel_err = abs(vdot_ab_gpu - vdot_ab) / abs(vdot_ab)
+            assert rel_err < 1e-4, rel_err
 
 
 @memoize
@@ -455,7 +507,7 @@ def summarize_error(obtained, desired, orig, thresh=1e-5):
             bad_count += 1
 
             if bad_count < bad_limit:
-                entries.append("%r (want: %r, got: %r, orig: %r)" % (
+                entries.append("{!r} (want: {!r}, got: {!r}, orig: {!r})".format(
                     obtained[i], desired[i], obtained[i], orig[i]))
         else:
             if bad_count:
@@ -468,6 +520,7 @@ def summarize_error(obtained, desired, orig, thresh=1e-5):
 
     return " ".join(entries)
 
+
 scan_test_counts = [
     10,
     2 ** 8 - 1,
@@ -523,6 +576,38 @@ def test_scan(ctx_factory, dtype, scan_cls):
         collect()
 
 
+@pytest.mark.parametrize("scan_cls", (GenericScanKernel, GenericDebugScanKernel))
+def test_scan_with_vectorargs_with_offsets(ctx_factory, scan_cls):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.tools import VectorArg
+
+    knl = scan_cls(
+            context, float,
+            arguments=[
+                VectorArg(float, "input", with_offset=True),
+                VectorArg(int, "segment", with_offset=True),
+                ],
+            input_expr="input[i]",
+            is_segment_start_expr="segment[i]",
+            scan_expr="a+b", neutral="0",
+            output_statement="""
+                input[i] = item;
+                """)
+
+    n = 20
+
+    host_data = np.random.randint(0, 10, n).astype(float)
+    dev_data = cl.array.to_device(queue, host_data)
+    segment_data = np.zeros(n, dtype=int)
+    dev_segment_data = cl.array.to_device(queue, segment_data)
+
+    knl(dev_data, dev_segment_data)
+
+    assert (dev_data.get() == np.cumsum(host_data)).all()
+
+
 def test_copy_if(ctx_factory):
     from pytest import importorskip
     importorskip("mako")
@@ -608,7 +693,6 @@ def test_index_preservation(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    from pyopencl.scan import GenericScanKernel, GenericDebugScanKernel
     classes = [GenericScanKernel]
 
     dev = context.devices[0]
@@ -652,7 +736,6 @@ def test_segmented_scan(ctx_factory):
         else:
             output_statement = "out[i] = item"
 
-        from pyopencl.scan import GenericScanKernel
         knl = GenericScanKernel(context, dtype,
                 arguments="__global %s *ary, __global char *segflags, "
                     "__global %s *out" % (ctype, ctype),
@@ -732,7 +815,8 @@ def test_segmented_scan(ctx_factory):
             print("%d excl:%s done" % (n, is_exclusive))
 
 
-def test_sort(ctx_factory):
+@pytest.mark.parametrize("scan_kernel", [GenericScanKernel, GenericDebugScanKernel])
+def test_sort(ctx_factory, scan_kernel):
     from pytest import importorskip
     importorskip("mako")
 
@@ -743,7 +827,7 @@ def test_sort(ctx_factory):
 
     from pyopencl.algorithm import RadixSort
     sort = RadixSort(context, "int *ary", key_expr="ary[i]",
-            sort_arg_names=["ary"])
+            sort_arg_names=["ary"], scan_kernel=scan_kernel)
 
     from pyopencl.clrandom import RanluxGenerator
     rng = RanluxGenerator(queue, seed=15)
@@ -752,6 +836,9 @@ def test_sort(ctx_factory):
 
     # intermediate arrays for largest size cause out-of-memory on low-end GPUs
     for n in scan_test_counts[:-1]:
+        if n >= 2000 and isinstance(scan_kernel, GenericDebugScanKernel):
+            continue
+
         print(n)
 
         print("  rng")
@@ -769,7 +856,7 @@ def test_sort(ctx_factory):
 
         numpy_elapsed = numpy_end-dev_end
         dev_elapsed = dev_end-dev_start
-        print ("  dev: %.2f MKeys/s numpy: %.2f MKeys/s ratio: %.2fx" % (
+        print("  dev: {:.2f} MKeys/s numpy: {:.2f} MKeys/s ratio: {:.2f}x".format(
                 1e-6*n/dev_elapsed, 1e-6*n/numpy_elapsed, numpy_elapsed/dev_elapsed))
         assert (a_dev_sorted.get() == a_sorted).all()
 
@@ -800,6 +887,104 @@ def test_list_builder(ctx_factory):
     assert (inf.lists.get()[-6:] == [1, 2, 2, 3, 3, 3]).all()
 
 
+def test_list_builder_with_memoryobject(ctx_factory):
+    from pytest import importorskip
+    importorskip("mako")
+
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.algorithm import ListOfListsBuilder
+    from pyopencl.tools import VectorArg
+    builder = ListOfListsBuilder(context, [("mylist", np.int32)], """//CL//
+            void generate(LIST_ARG_DECL USER_ARG_DECL index_type i)
+            {
+                APPEND_mylist(input_list[i]);
+            }
+            """, arg_decls=[VectorArg(float, "input_list")])
+
+    n = 10000
+    input_list = cl.array.zeros(queue, (n,), float)
+    result, evt = builder(queue, n, input_list.data)
+
+    inf = result["mylist"]
+    assert inf.count == n
+    assert (inf.lists.get() == 0).all()
+
+
+def test_list_builder_with_offset(ctx_factory):
+    from pytest import importorskip
+    importorskip("mako")
+
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.algorithm import ListOfListsBuilder
+    from pyopencl.tools import VectorArg
+    builder = ListOfListsBuilder(context, [("mylist", np.int32)], """//CL//
+            void generate(LIST_ARG_DECL USER_ARG_DECL index_type i)
+            {
+                APPEND_mylist(input_list[i]);
+            }
+            """, arg_decls=[
+                VectorArg(float, "input_list", with_offset=True)])
+
+    n = 10000
+    input_list = cl.array.zeros(queue, (n + 10,), float)
+    input_list[10:] = 1
+
+    result, evt = builder(queue, n, input_list[10:])
+
+    inf = result["mylist"]
+    assert inf.count == n
+    assert (inf.lists.get() == 1).all()
+
+
+def test_list_builder_with_empty_elim(ctx_factory):
+    from pytest import importorskip
+    importorskip("mako")
+
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.algorithm import ListOfListsBuilder
+
+    builder = ListOfListsBuilder(
+        context,
+        [("mylist1", np.int32), ("mylist2", np.int32), ("mylist3", np.int32)],
+        """//CL//
+            void generate(LIST_ARG_DECL USER_ARG_DECL index_type i)
+            {
+                if (i % 5 == 0)
+                {
+                    for (int j = 0; j < i / 5; ++j)
+                    {
+                        APPEND_mylist1(j);
+                        APPEND_mylist2(j + 1);
+                        APPEND_mylist3(j);
+                    }
+                }
+            }
+        """,
+        arg_decls=[],
+        eliminate_empty_output_lists=["mylist1", "mylist2"])
+
+    result, evt = builder(queue, 1000)
+
+    mylist1 = result["mylist1"]
+    assert mylist1.count == 19900
+    assert (mylist1.starts.get()[:5] == [0, 1, 3, 6, 10]).all()
+    assert (mylist1.nonempty_indices.get()[:5] == [5, 10, 15, 20, 25]).all()
+    assert (mylist1.lists.get()[:6] == [0, 0, 1, 0, 1, 2]).all()
+    mylist2 = result["mylist2"]
+    assert mylist2.count == 19900
+    assert (mylist2.lists.get()[:6] == [1, 1, 2, 1, 2, 3]).all()
+    mylist3 = result["mylist3"]
+    assert mylist3.count == 19900
+    assert (mylist3.starts.get()[:10] == [0, 0, 0, 0, 0, 0, 1, 1, 1, 1]).all()
+    assert (mylist3.lists.get()[:6] == [0, 0, 1, 0, 1, 2]).all()
+
+
 def test_key_value_sorter(ctx_factory):
     from pytest import importorskip
     importorskip("mako")
@@ -833,11 +1018,130 @@ def test_key_value_sorter(ctx_factory):
 # }}}
 
 
+# {{{ bitonic sort
+
+@pytest.mark.parametrize("size", [
+    512,
+    4,
+    16
+    ])
+@pytest.mark.parametrize("dtype", [
+    np.int32,
+    np.float32,
+    np.float64
+    ])
+@pytest.mark.bitonic
+def test_bitonic_sort(ctx_factory, size, dtype):
+    ctx = cl.create_some_context()
+    queue = cl.CommandQueue(ctx)
+
+    dev = ctx.devices[0]
+    if (dev.platform.name == "Apple" and dev.type & cl.device_type.CPU):
+        pytest.xfail("Bitonic sort won't work on Apple CPU: no workgroup "
+            "parallelism")
+    if (dev.platform.name == "Portable Computing Language"
+            and dtype == np.float64
+            and get_pocl_version(dev.platform) < (1, 0)):
+        pytest.xfail("Double precision bitonic sort doesn't work on POCL < 1.0")
+
+    if dtype == np.float64 and not has_double_support(dev):
+        from pytest import skip
+        skip("double precision not supported on %s" % dev)
+
+    import pyopencl.clrandom as clrandom
+    from pyopencl.bitonic_sort import BitonicSort
+
+    s = clrandom.rand(queue, (2, size, 3,), dtype, luxury=None, a=0, b=239482333)
+    sgs = s.copy()
+    # enqueue_marker crashes under CL 1.1 pocl if there is anything to wait for
+    # (no clEnqueueWaitForEvents) https://github.com/inducer/pyopencl/pull/237
+    if (dev.platform.name == "Portable Computing Language"
+            and cl.get_cl_header_version() < (1, 2)):
+        sgs.finish()
+    sorter = BitonicSort(ctx)
+    sgs, evt = sorter(sgs, axis=1)
+    assert np.array_equal(np.sort(s.get(), axis=1), sgs.get())
+
+
+@pytest.mark.parametrize("size", [
+    0,
+    4,
+    2**14,
+    2**18,
+    ])
+@pytest.mark.parametrize("dtype", [
+    np.int32,
+    np.float32,
+    np.float64
+    ])
+@pytest.mark.bitonic
+def test_bitonic_argsort(ctx_factory, size, dtype):
+    import sys
+    is_pypy = "__pypy__" in sys.builtin_module_names
+
+    if not size and is_pypy:
+        # https://bitbucket.org/pypy/numpy/issues/53/specifying-strides-on-zero-sized-array
+        pytest.xfail("pypy doesn't seem to handle as_strided "
+                "on zero-sized arrays very well")
+
+    ctx = cl.create_some_context()
+    queue = cl.CommandQueue(ctx)
+
+    device = queue.device
+    if device.platform.vendor == "The pocl project" \
+            and device.type & cl.device_type.GPU:
+        pytest.xfail("bitonic argsort fails on POCL + Nvidia,"
+                "at least the K40, as of pocl 1.6, 2021-01-20")
+
+    dev = ctx.devices[0]
+    if (dev.platform.name == "Portable Computing Language"
+            and sys.platform == "darwin"):
+        pytest.xfail("Bitonic sort crashes on Apple POCL")
+    if (dev.platform.name == "Apple" and dev.type & cl.device_type.CPU):
+        pytest.xfail("Bitonic sort won't work on Apple CPU: no workgroup "
+            "parallelism")
+    if (dev.platform.name == "Portable Computing Language"
+            and dtype == np.float64
+            and get_pocl_version(dev.platform) < (1, 0)):
+        pytest.xfail("Double precision bitonic sort doesn't work on POCL < 1.0")
+
+    if dtype == np.float64 and not has_double_support(dev):
+        from pytest import skip
+        skip("double precision not supported on %s" % dev)
+
+    import pyopencl.clrandom as clrandom
+    from pyopencl.bitonic_sort import BitonicSort
+
+    index = cl_array.arange(queue, 0, size, 1, dtype=np.int32)
+    m = clrandom.rand(queue, (size,), dtype, luxury=None, a=0, b=239432234)
+
+    sorterm = BitonicSort(ctx)
+
+    ms = m.copy()
+    # enqueue_marker crashes under CL 1.1 pocl if there is anything to wait for
+    # (no clEnqueueWaitForEvents) https://github.com/inducer/pyopencl/pull/237
+    if (dev.platform.name == "Portable Computing Language"
+            and cl.get_cl_header_version() < (1, 2)):
+        ms.finish()
+        index.finish()
+    ms, evt = sorterm(ms, idx=index, axis=0)
+
+    assert np.array_equal(np.sort(m.get()), ms.get())
+
+    # may be False because of identical values in array
+    # assert np.array_equal(np.argsort(m.get()), index.get())
+
+    # Check values by indices
+    assert np.array_equal(m.get()[np.argsort(m.get())], m.get()[index.get()])
+
+# }}}
+
+
 if __name__ == "__main__":
     if len(sys.argv) > 1:
         exec(sys.argv[1])
     else:
-        from py.test.cmdline import main
+        from pytest import main
         main([__file__])
 
 # vim: filetype=pyopencl:fdm=marker
