@@ -1,5 +1,3 @@
-from __future__ import division, absolute_import, print_function
-
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
 __license__ = """
@@ -22,7 +20,9 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-from six.moves import range
+# avoid spurious: pytest.mark.parametrize is not callable
+# pylint: disable=not-callable
+
 
 import numpy as np
 import numpy.linalg as la
@@ -30,8 +30,12 @@ import pytest
 
 import pyopencl as cl
 import pyopencl.array as cl_array
+import pyopencl.cltypes as cltypes
+import pyopencl.clrandom
 from pyopencl.tools import (  # noqa
-        pytest_generate_tests_for_pyopencl as pytest_generate_tests)
+        pytest_generate_tests_for_pyopencl as pytest_generate_tests,
+        ImmediateAllocator, DeferredAllocator)
+from pyopencl.characterize import get_pocl_version
 
 # Are CL implementations crashy? You be the judge. :)
 try:
@@ -42,10 +46,17 @@ else:
     faulthandler.enable()
 
 
-def _skip_if_pocl(plat, msg='unsupported by pocl'):
+def _xfail_if_pocl(plat, up_to_version, msg="unsupported by pocl"):
     if plat.vendor == "The pocl project":
-        import pytest
-        pytest.skip(msg)
+        if up_to_version is None or get_pocl_version(plat) <= up_to_version:
+            pytest.xfail(msg)
+
+
+def _xfail_if_pocl_gpu(device, what):
+    if device.platform.vendor == "The pocl project" \
+            and device.type & cl.device_type.GPU:
+        pytest.xfail(f"POCL's {what} support don't work right on Nvidia GPUs, "
+                "at least the Titan V, as of pocl 1.6, 2021-01-20")
 
 
 def test_get_info(ctx_factory):
@@ -53,6 +64,9 @@ def test_get_info(ctx_factory):
     device, = ctx.devices
     platform = device.platform
 
+    device.persistent_unique_id
+    device.hashable_model_and_version_identifier
+
     failure_count = [0]
 
     pocl_quirks = [
@@ -117,8 +131,7 @@ def test_get_info(ctx_factory):
 
     def do_test(cl_obj, info_cls, func=None, try_attr_form=True):
         if func is None:
-            def func(info):
-                cl_obj.get_info(info)
+            func = cl_obj.get_info
 
         for info_name in dir(info_cls):
             if not info_name.startswith("_") and info_name != "to_string":
@@ -132,7 +145,7 @@ def test_get_info(ctx_factory):
 
                 try:
                     func(info)
-                except:
+                except Exception:
                     msg = "failed get_info", type(cl_obj), info_name
 
                     if find_quirk(QUIRKS, cl_obj, info):
@@ -143,7 +156,7 @@ def test_get_info(ctx_factory):
                 if try_attr_form:
                     try:
                         getattr(cl_obj, info_name.lower())
-                    except:
+                    except Exception:
                         print("failed attr-based get_info", type(cl_obj), info_name)
 
                         if find_quirk(QUIRKS, cl_obj, info):
@@ -178,9 +191,20 @@ def test_get_info(ctx_factory):
 
     do_test(a_buf, cl.mem_info)
 
-    kernel = prg.sum
+    kernel = prg.all_kernels()[0]
     do_test(kernel, cl.kernel_info)
 
+    for i in range(2):  # exercise cache
+        for info_name in dir(cl.kernel_work_group_info):
+            if not info_name.startswith("_") and info_name != "to_string":
+                try:
+                    print("kernel_wg_info: %s" % info_name)
+                    kernel.get_work_group_info(
+                            getattr(cl.kernel_work_group_info, info_name),
+                            device)
+                except cl.LogicError as err:
+                    print("<error: %s>" % err)
+
     evt = kernel(queue, (n,), None, a_buf)
     do_test(evt, cl.event_info)
 
@@ -291,7 +315,21 @@ def test_image_format_constructor():
 
     assert iform.channel_order == cl.channel_order.RGBA
     assert iform.channel_data_type == cl.channel_type.FLOAT
-    assert not iform.__dict__
+
+    if not cl._PYPY:
+        assert not hasattr(iform, "__dict__")
+
+
+def test_device_topology_amd_constructor():
+    # doesn't need cl_amd_device_attribute_query support to succeed
+    topol = cl.DeviceTopologyAmd(3, 4, 5)
+
+    assert topol.bus == 3
+    assert topol.device == 4
+    assert topol.function == 5
+
+    if not cl._PYPY:
+        assert not hasattr(topol, "__dict__")
 
 
 def test_nonempty_supported_image_formats(ctx_factory):
@@ -336,7 +374,7 @@ def test_that_python_args_fail(ctx_factory):
     prg.mult(queue, a.shape, None, a_buf, np.float32(2), np.int32(3))
 
     a_result = np.empty_like(a)
-    cl.enqueue_read_buffer(queue, a_buf, a_result).wait()
+    cl.enqueue_copy(queue, a_buf, a_result).wait()
 
 
 def test_image_2d(ctx_factory):
@@ -351,7 +389,7 @@ def test_image_2d(ctx_factory):
     if "Intel" in device.vendor and "31360.31426" in device.version:
         from pytest import skip
         skip("images crashy on %s" % device)
-    _skip_if_pocl(device.platform, 'pocl does not support CL_ADDRESS_CLAMP')
+    _xfail_if_pocl(device.platform, None, "pocl does not support CL_ADDRESS_CLAMP")
 
     prg = cl.Program(context, """
         __kernel void copy_image(
@@ -423,7 +461,7 @@ def test_image_3d(ctx_factory):
     if device.platform.vendor == "Intel(R) Corporation":
         from pytest import skip
         skip("images crashy on %s" % device)
-    _skip_if_pocl(device.platform, 'pocl does not support CL_ADDRESS_CLAMP')
+    _xfail_if_pocl(device.platform, None, "pocl does not support CL_ADDRESS_CLAMP")
 
     prg = cl.Program(context, """
         __kernel void copy_image_plane(
@@ -498,8 +536,8 @@ def test_copy_buffer(ctx_factory):
     buf1 = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
     buf2 = cl.Buffer(context, mf.WRITE_ONLY, b.nbytes)
 
-    cl.enqueue_copy_buffer(queue, buf1, buf2).wait()
-    cl.enqueue_read_buffer(queue, buf2, b).wait()
+    cl.enqueue_copy(queue, buf2, buf1).wait()
+    cl.enqueue_copy(queue, b, buf2).wait()
 
     assert la.norm(a - b) == 0
 
@@ -524,20 +562,42 @@ def test_mempool(ctx_factory):
     pool.stop_holding()
 
 
-def test_mempool_2():
-    from pyopencl.tools import MemoryPool
+def test_mempool_2(ctx_factory):
+    from pyopencl.tools import MemoryPool, ImmediateAllocator
     from random import randrange
 
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    pool = MemoryPool(ImmediateAllocator(queue))
+
     for i in range(2000):
         s = randrange(1 << 31) >> randrange(32)
-        bin_nr = MemoryPool.bin_number(s)
-        asize = MemoryPool.alloc_size(bin_nr)
+        bin_nr = pool.bin_number(s)
+        asize = pool.alloc_size(bin_nr)
 
         assert asize >= s, s
-        assert MemoryPool.bin_number(asize) == bin_nr, s
+        assert pool.bin_number(asize) == bin_nr, s
         assert asize < asize*(1+1/8)
 
 
+@pytest.mark.parametrize("allocator_cls", [ImmediateAllocator, DeferredAllocator])
+def test_allocator(ctx_factory, allocator_cls):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    if allocator_cls is DeferredAllocator:
+        allocator = allocator_cls(context)
+    else:
+        allocator = allocator_cls(queue)
+
+    mem = allocator(15)
+    mem2 = allocator(0)
+
+    assert mem is not None
+    assert mem2 is None
+
+
 def test_vector_args(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -547,14 +607,14 @@ def test_vector_args(ctx_factory):
         { dest[get_global_id(0)] = x; }
         """).build()
 
-    x = cl_array.vec.make_float4(1, 2, 3, 4)
-    dest = np.empty(50000, cl_array.vec.float4)
+    x = cltypes.make_float4(1, 2, 3, 4)
+    dest = np.empty(50000, cltypes.float4)
     mf = cl.mem_flags
     dest_buf = cl.Buffer(context, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=dest)
 
     prg.set_vec(queue, dest.shape, None, x, dest_buf)
 
-    cl.enqueue_read_buffer(queue, dest_buf, dest).wait()
+    cl.enqueue_copy(queue, dest, dest_buf).wait()
 
     assert (dest == x).all()
 
@@ -596,9 +656,11 @@ def test_context_dep_memoize(ctx_factory):
     assert counter[0] == 1
 
 
-def test_can_build_binary(ctx_factory):
+def test_can_build_and_run_binary(ctx_factory):
     ctx = ctx_factory()
-    device, = ctx.devices
+    queue = cl.CommandQueue(ctx)
+
+    device = queue.device
 
     program = cl.Program(ctx, """
     __kernel void simple(__global float *in, __global float *out)
@@ -611,11 +673,25 @@ def test_can_build_binary(ctx_factory):
     foo = cl.Program(ctx, [device], [binary])
     foo.build()
 
+    n = 256
+    a_dev = cl.clrandom.rand(queue, n, np.float32)
+    dest_dev = cl_array.empty_like(a_dev)
+
+    foo.simple(queue, (n,), (16,), a_dev.data, dest_dev.data)
+
 
 def test_enqueue_barrier_marker(ctx_factory):
     ctx = ctx_factory()
-    _skip_if_pocl(ctx.devices[0].platform, 'pocl crashes on enqueue_barrier')
+    # Still relevant on pocl 1.0RC1.
+    _xfail_if_pocl(
+            ctx.devices[0].platform, (1, 0), "pocl crashes on enqueue_barrier")
+
     queue = cl.CommandQueue(ctx)
+
+    if queue._get_cl_version() >= (1, 2) and cl.get_cl_header_version() <= (1, 1):
+        pytest.skip("CL impl version >= 1.2, header version <= 1.1--cannot be sure "
+                "that clEnqueueWaitForEvents is implemented")
+
     cl.enqueue_barrier(queue)
     evt1 = cl.enqueue_marker(queue)
     evt2 = cl.enqueue_marker(queue, wait_for=[evt1])
@@ -631,47 +707,17 @@ def test_wait_for_events(ctx_factory):
 
 
 def test_unload_compiler(platform):
-    if (platform._get_cl_version() < (1, 2) or
-            cl.get_cl_header_version() < (1, 2)):
+    if (platform._get_cl_version() < (1, 2)
+            or cl.get_cl_header_version() < (1, 2)):
         from pytest import skip
         skip("clUnloadPlatformCompiler is only available in OpenCL 1.2")
-    _skip_if_pocl(platform, 'pocl does not support unloading compiler')
+    _xfail_if_pocl(platform, (0, 13), "pocl does not support unloading compiler")
     if platform.vendor == "Intel(R) Corporation":
         from pytest import skip
         skip("Intel proprietary driver does not support unloading compiler")
     cl.unload_platform_compiler(platform)
 
 
-def test_enqueue_task(ctx_factory):
-    ctx = ctx_factory()
-    queue = cl.CommandQueue(ctx)
-    mf = cl.mem_flags
-
-    prg = cl.Program(ctx, """
-    __kernel void
-    reverse(__global const float *in, __global float *out, int n)
-    {
-        for (int i = 0;i < n;i++) {
-            out[i] = in[n - 1 - i];
-        }
-    }
-    """).build()
-    knl = prg.reverse
-
-    n = 100
-    a = np.random.rand(n).astype(np.float32)
-    b = np.empty_like(a)
-
-    buf1 = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
-    buf2 = cl.Buffer(ctx, mf.WRITE_ONLY, b.nbytes)
-
-    knl.set_args(buf1, buf2, np.int32(n))
-    cl.enqueue_task(queue, knl)
-
-    cl.enqueue_copy(queue, b, buf2).wait()
-    assert la.norm(a[::-1] - b) == 0
-
-
 def test_platform_get_devices(ctx_factory):
     ctx = ctx_factory()
     platform = ctx.devices[0].platform
@@ -682,8 +728,8 @@ def test_platform_get_devices(ctx_factory):
 
     dev_types = [cl.device_type.ACCELERATOR, cl.device_type.ALL,
                  cl.device_type.CPU, cl.device_type.DEFAULT, cl.device_type.GPU]
-    if (platform._get_cl_version() >= (1, 2) and
-            cl.get_cl_header_version() >= (1, 2)
+    if (platform._get_cl_version() >= (1, 2)
+            and cl.get_cl_header_version() >= (1, 2)
             and not platform.name.lower().startswith("nvidia")):
         dev_types.append(cl.device_type.CUSTOM)
 
@@ -692,7 +738,7 @@ def test_platform_get_devices(ctx_factory):
         devs = platform.get_devices(dev_type)
         if dev_type in (cl.device_type.DEFAULT,
                         cl.device_type.ALL,
-                        getattr(cl.device_type, 'CUSTOM', None)):
+                        getattr(cl.device_type, "CUSTOM", None)):
             continue
         for dev in devs:
             assert dev.type & dev_type == dev_type
@@ -700,14 +746,14 @@ def test_platform_get_devices(ctx_factory):
 
 def test_user_event(ctx_factory):
     ctx = ctx_factory()
-    if (ctx._get_cl_version() < (1, 1) and
-            cl.get_cl_header_version() < (1, 1)):
+    if (ctx._get_cl_version() < (1, 1)
+            and cl.get_cl_header_version() < (1, 1)):
         from pytest import skip
         skip("UserEvent is only available in OpenCL 1.1")
 
-    if ctx.devices[0].platform.name == "Portable Computing Language":
-        # https://github.com/pocl/pocl/issues/201
-        pytest.xfail("POCL's user events don't work right")
+    # https://github.com/pocl/pocl/issues/201
+    _xfail_if_pocl(ctx.devices[0].platform, (0, 13),
+            "pocl's user events don't work right")
 
     status = {}
 
@@ -725,26 +771,30 @@ def test_user_event(ctx_factory):
     Thread(target=event_waiter1, args=(evt, 1)).start()
     sleep(.05)
     if status.get(1, False):
-        raise RuntimeError('UserEvent triggered before set_status')
+        raise RuntimeError("UserEvent triggered before set_status")
     evt.set_status(cl.command_execution_status.COMPLETE)
     sleep(.05)
     if not status.get(1, False):
-        raise RuntimeError('UserEvent.wait timeout')
+        raise RuntimeError("UserEvent.wait timeout")
     assert evt.command_execution_status == cl.command_execution_status.COMPLETE
 
     evt = cl.UserEvent(ctx)
     Thread(target=event_waiter2, args=(evt, 2)).start()
     sleep(.05)
     if status.get(2, False):
-        raise RuntimeError('UserEvent triggered before set_status')
+        raise RuntimeError("UserEvent triggered before set_status")
     evt.set_status(cl.command_execution_status.COMPLETE)
     sleep(.05)
     if not status.get(2, False):
-        raise RuntimeError('cl.wait_for_events timeout on UserEvent')
+        raise RuntimeError("cl.wait_for_events timeout on UserEvent")
     assert evt.command_execution_status == cl.command_execution_status.COMPLETE
 
 
 def test_buffer_get_host_array(ctx_factory):
+    if cl._PYPY:
+        # FIXME
+        pytest.xfail("Buffer.get_host_array not yet working on pypy")
+
     ctx = ctx_factory()
     mf = cl.mem_flags
 
@@ -752,8 +802,8 @@ def test_buffer_get_host_array(ctx_factory):
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.USE_HOST_PTR, hostbuf=host_buf)
     host_buf2 = buf.get_host_array(25, np.float32)
     assert (host_buf == host_buf2).all()
-    assert (host_buf.__array_interface__['data'][0] ==
-            host_buf.__array_interface__['data'][0])
+    assert (host_buf.__array_interface__["data"][0]
+            == host_buf.__array_interface__["data"][0])
     assert host_buf2.base is buf
 
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.ALLOC_HOST_PTR, size=100)
@@ -792,11 +842,17 @@ def test_program_valued_get_info(ctx_factory):
 
 
 def test_event_set_callback(ctx_factory):
+    import sys
+    if sys.platform.startswith("win"):
+        pytest.xfail("Event.set_callback not present on Windows")
+
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
 
+    _xfail_if_pocl_gpu(queue.device, "event callbacks")
+
     if ctx._get_cl_version() < (1, 1):
-        pytest.skip("OpenCL 1.1 or newer required fro set_callback")
+        pytest.skip("OpenCL 1.1 or newer required for set_callback")
 
     a_np = np.random.rand(50000).astype(np.float32)
     b_np = np.random.rand(50000).astype(np.float32)
@@ -830,9 +886,17 @@ def test_event_set_callback(ctx_factory):
 
     queue.finish()
 
+    counter = 0
+
     # yuck
-    from time import sleep
-    sleep(0.1)
+    while not got_called:
+        from time import sleep
+        sleep(0.01)
+
+        # wait up to five seconds (?!)
+        counter += 1
+        if counter >= 500:
+            break
 
     assert got_called
 
@@ -841,6 +905,8 @@ def test_global_offset(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
+    _xfail_if_pocl_gpu(queue.device, "global offset")
+
     prg = cl.Program(context, """
         __kernel void mult(__global float *a)
         { a[get_global_id(0)] *= 2; }
@@ -863,6 +929,298 @@ def test_global_offset(ctx_factory):
     assert (a_2 == 2*a).all()
 
 
+def test_sub_buffers(ctx_factory):
+    ctx = ctx_factory()
+    if (ctx._get_cl_version() < (1, 1)
+            or cl.get_cl_header_version() < (1, 1)):
+        from pytest import skip
+        skip("sub-buffers are only available in OpenCL 1.1")
+
+    alignment = ctx.devices[0].mem_base_addr_align
+
+    queue = cl.CommandQueue(ctx)
+
+    n = 30000
+    a = (np.random.rand(n) * 100).astype(np.uint8)
+
+    mf = cl.mem_flags
+    a_buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
+
+    start = (5000 // alignment) * alignment
+    stop = start + 20 * alignment
+
+    a_sub_ref = a[start:stop]
+
+    a_sub = np.empty_like(a_sub_ref)
+    cl.enqueue_copy(queue, a_sub, a_buf[start:stop])
+
+    assert np.array_equal(a_sub, a_sub_ref)
+
+
+def test_spirv(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    if (ctx._get_cl_version() < (2, 1)
+            or cl.get_cl_header_version() < (2, 1)):
+        pytest.skip("SPIR-V program creation only available "
+                "in OpenCL 2.1 and higher")
+
+    n = 50000
+
+    a_dev = cl.clrandom.rand(queue, n, np.float32)
+    b_dev = cl.clrandom.rand(queue, n, np.float32)
+    dest_dev = cl_array.empty_like(a_dev)
+
+    with open("add-vectors-%d.spv" % queue.device.address_bits, "rb") as spv_file:
+        spv = spv_file.read()
+
+    prg = cl.Program(ctx, spv).build()
+    if (not prg.all_kernels()
+            and queue.device.platform.name.startswith("AMD Accelerated")):
+        pytest.skip("SPIR-V program creation on AMD did not result in any kernels")
+
+    prg.sum(queue, a_dev.shape, None, a_dev.data, b_dev.data, dest_dev.data)
+
+    assert la.norm((dest_dev - (a_dev+b_dev)).get()) < 1e-7
+
+
+def test_coarse_grain_svm(ctx_factory):
+    import sys
+    is_pypy = "__pypy__" in sys.builtin_module_names
+
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    _xfail_if_pocl_gpu(queue.device, "SVM")
+
+    dev = ctx.devices[0]
+
+    from pyopencl.characterize import has_coarse_grain_buffer_svm
+    from pytest import skip
+    if not has_coarse_grain_buffer_svm(queue.device):
+        skip("device does not support coarse-grain SVM")
+
+    if ("AMD" in dev.platform.name
+            and dev.type & cl.device_type.CPU):
+        pytest.xfail("AMD CPU doesn't do coarse-grain SVM")
+    if ("AMD" in dev.platform.name
+            and dev.type & cl.device_type.GPU):
+        pytest.xfail("AMD GPU crashes on SVM unmap")
+
+    n = 3000
+    svm_ary = cl.SVM(cl.csvm_empty(ctx, (n,), np.float32, alignment=64))
+    if not is_pypy:
+        # https://bitbucket.org/pypy/numpy/issues/52
+        assert isinstance(svm_ary.mem.base, cl.SVMAllocation)
+
+    cl.enqueue_svm_memfill(queue, svm_ary, np.zeros((), svm_ary.mem.dtype))
+
+    with svm_ary.map_rw(queue) as ary:
+        ary.fill(17)
+        orig_ary = ary.copy()
+
+    prg = cl.Program(ctx, """
+        __kernel void twice(__global float *a_g)
+        {
+          a_g[get_global_id(0)] *= 2;
+        }
+        """).build()
+
+    prg.twice(queue, svm_ary.mem.shape, None, svm_ary)
+
+    with svm_ary.map_ro(queue) as ary:
+        print(ary)
+        assert np.array_equal(orig_ary*2, ary)
+
+    new_ary = np.empty_like(orig_ary)
+    new_ary.fill(-1)
+
+    if ctx.devices[0].platform.name != "Portable Computing Language":
+        # "Blocking memcpy is unimplemented (clEnqueueSVMMemcpy.c:61)"
+        # in pocl up to and including 1.0rc1.
+
+        cl.enqueue_copy(queue, new_ary, svm_ary)
+        assert np.array_equal(orig_ary*2, new_ary)
+
+    # {{{ https://github.com/inducer/pyopencl/issues/372
+
+    buf_arr = cl.svm_empty(ctx, cl.svm_mem_flags.READ_ONLY, 10, np.int32)
+    out_arr = cl.svm_empty(ctx, cl.svm_mem_flags.READ_WRITE, 10, np.int32)
+
+    svm_buf_arr = cl.SVM(buf_arr)
+    svm_out_arr = cl.SVM(out_arr)
+    with svm_buf_arr.map_rw(queue) as ary:
+        ary.fill(17)
+
+    prg_ro = cl.Program(ctx, r"""
+        __kernel void twice_ro(__global int *out_g, __global int *in_g)
+        {
+          out_g[get_global_id(0)] = 2*in_g[get_global_id(0)];
+        }
+        """).build()
+
+    prg_ro.twice_ro(queue, buf_arr.shape, None, svm_out_arr, svm_buf_arr)
+
+    with svm_out_arr.map_ro(queue) as ary:
+        print(ary)
+
+    # }}}
+
+
+def test_fine_grain_svm(ctx_factory):
+    import sys
+    is_pypy = "__pypy__" in sys.builtin_module_names
+
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    _xfail_if_pocl_gpu(queue.device, "GPU SVM")
+
+    from pyopencl.characterize import has_fine_grain_buffer_svm
+    from pytest import skip
+    if not has_fine_grain_buffer_svm(queue.device):
+        skip("device does not support fine-grain SVM")
+
+    n = 3000
+    ary = cl.fsvm_empty(ctx, n, np.float32, alignment=64)
+
+    if not is_pypy:
+        # https://bitbucket.org/pypy/numpy/issues/52
+        assert isinstance(ary.base, cl.SVMAllocation)
+
+    ary.fill(17)
+    orig_ary = ary.copy()
+
+    prg = cl.Program(ctx, """
+        __kernel void twice(__global float *a_g)
+        {
+          a_g[get_global_id(0)] *= 2;
+        }
+        """).build()
+
+    prg.twice(queue, ary.shape, None, cl.SVM(ary))
+    queue.finish()
+
+    print(ary)
+    assert np.array_equal(orig_ary*2, ary)
+
+
+@pytest.mark.parametrize("dtype", [
+    np.uint,
+    cltypes.uint2,
+    ])
+def test_map_dtype(ctx_factory, dtype):
+    if cl._PYPY:
+        # FIXME
+        pytest.xfail("enqueue_map_buffer not yet working on pypy")
+
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    dt = np.dtype(dtype)
+
+    b = pyopencl.Buffer(ctx,
+                        pyopencl.mem_flags.READ_ONLY,
+                        dt.itemsize)
+    array, ev = pyopencl.enqueue_map_buffer(queue, b, pyopencl.map_flags.WRITE, 0,
+                                            (1,), dt)
+    with array.base:
+        print(array.dtype)
+        assert array.dtype == dt
+
+
+def test_compile_link(ctx_factory):
+    ctx = ctx_factory()
+
+    if ctx._get_cl_version() < (1, 2) or cl.get_cl_header_version() < (1, 2):
+        pytest.skip("Context and ICD loader must understand CL1.2 for compile/link")
+
+    platform = ctx.devices[0].platform
+    if platform.name == "Apple":
+        pytest.skip("Apple doesn't like our compile/link test")
+
+    queue = cl.CommandQueue(ctx)
+    vsink_prg = cl.Program(ctx, """//CL//
+        void value_sink(float x)
+        {
+        }
+        """).compile()
+    main_prg = cl.Program(ctx, """//CL//
+        void value_sink(float x);
+
+        __kernel void experiment()
+        {
+            value_sink(3.1415f + get_global_id(0));
+        }
+        """).compile()
+    z = cl.link_program(ctx, [vsink_prg, main_prg], devices=ctx.devices)
+    z.experiment(queue, (128**2,), (128,))
+    queue.finish()
+
+
+def test_copy_buffer_rect(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    _xfail_if_pocl_gpu(queue.device, "rectangular copies")
+
+    arr1 = cl_array.zeros(queue, (2, 3), "f")
+    arr2 = cl_array.zeros(queue, (4, 5), "f")
+    arr1.fill(1)
+    cl.enqueue_copy(
+            queue, arr2.data, arr1.data,
+            src_origin=(0, 0), dst_origin=(1, 1),
+            region=arr1.shape[::-1])
+
+
+def test_threaded_nanny_events(ctx_factory):
+    # https://github.com/inducer/pyopencl/issues/296
+
+    import gc
+    import threading
+
+    def create_arrays_thread(n1=10, n2=20):
+        ctx = ctx_factory()
+        queue = cl.CommandQueue(ctx)
+        for i1 in range(n2):
+            for i in range(n1):
+                acl = cl.array.zeros(queue, 10, dtype=np.float32)
+                acl.get()
+            # Garbage collection triggers the error
+            print("collected ", str(gc.collect()))
+            print("stats ", gc.get_stats())
+
+    t1 = threading.Thread(target=create_arrays_thread)
+    t2 = threading.Thread(target=create_arrays_thread)
+
+    t1.start()
+    t2.start()
+
+    t1.join()
+    t2.join()
+
+
+@pytest.mark.parametrize("empty_shape", [(0,), (3, 0, 2)])
+def test_empty_ndrange(ctx_factory, empty_shape):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    if ctx._get_cl_version() < (1, 2) or cl.get_cl_header_version() < (1, 2):
+        pytest.skip("OpenCL 1.2 required for empty NDRange suuport")
+
+    a = cl_array.zeros(queue, empty_shape, dtype=np.float32)
+
+    prg = cl.Program(ctx, """
+        __kernel void add_two(__global float *a_g)
+        {
+          a_g[get_global_id(0)] += 2;
+        }
+        """).build()
+
+    prg.add_two(queue, a.shape, None, a.data, allow_empty_ndrange=True)
+
+
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
     import pyopencl  # noqa
@@ -871,5 +1229,5 @@ if __name__ == "__main__":
     if len(sys.argv) > 1:
         exec(sys.argv[1])
     else:
-        from py.test.cmdline import main
+        from pytest import main
         main([__file__])
